# SPDX-FileCopyrightText: skforecast team
# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later AND BSD-3-Clause

import pandas as pd
from typing import Union
import numpy as np


class WeightFunction:
    """Callable class for sample weights that can be pickled.

    This class wraps the weights_series and provides a callable interface
    compatible with ForecasterRecursive's weight_func parameter. Unlike
    local functions with closures, instances of this class can be pickled
    using standard pickle/joblib.

    Args:
        weights_series: Series containing weight values for each index.

    Examples:
        >>> import pandas as pd
        >>> import pickle
        >>> weights = pd.Series([1.0, 0.9, 0.8], index=[0, 1, 2])
        >>> weight_func = WeightFunction(weights)
        >>> weight_func(pd.Index([0, 1]))
        array([1. , 0.9])
        >>> # Can be pickled
        >>> pickled = pickle.dumps(weight_func)
        >>> unpickled = pickle.loads(pickled)
        >>> unpickled(pd.Index([0, 1]))
        array([1. , 0.9])
    """

    def __init__(self, weights_series: pd.Series):
        """Initialize with a weights series.

        Args:
            weights_series: Series containing weight values for each index.
        """
        self.weights_series = weights_series

    def __call__(
        self, index: Union[pd.Index, np.ndarray, list]
    ) -> Union[float, np.ndarray]:
        """Return sample weights for given index.

        Args:
            index: Index or indices to get weights for.

        Returns:
            Weight value(s) corresponding to the index.
        """
        return custom_weights(index, self.weights_series)

    def __repr__(self):
        """String representation."""
        return f"WeightFunction(weights_series with {len(self.weights_series)} entries)"


def custom_weights(index, weights_series: pd.Series) -> float:
    """
    Return 0 if index is in or near any gap.

    Args:
        index (pd.Index):
            The index to check.
        weights_series (pd.Series):
            Series containing weights.

    Returns:
        float: The weight corresponding to the index.

    Examples:
        >>> from spotforecast2_safe.data.fetch_data import fetch_data
        >>> from spotforecast2_safe.preprocessing.imputation import custom_weights
        >>> data = fetch_data()
        >>> _, missing_weights = get_missing_weights(data, window_size=72, verbose=False)
        >>> for idx in data.index[:5]:
        ...     weight = custom_weights(idx, missing_weights)
        ...     print(f"Index: {idx}, Weight: {weight}")
    """
    # do plausibility check
    if isinstance(index, pd.Index):
        if not index.isin(weights_series.index).all():
            raise ValueError("Index not found in weights_series.")
        return weights_series.loc[index].values

    if index not in weights_series.index:
        raise ValueError("Index not found in weights_series.")
    return weights_series.loc[index]


def get_missing_weights(
    data: pd.DataFrame, window_size: int = 72, verbose: bool = False
) -> tuple[pd.DataFrame, pd.Series]:
    """
    Return imputed DataFrame and a series indicating missing weights.

    Args:
        data (pd.DataFrame):
            The input dataset.
        window_size (int):
            The size of the rolling window to consider for missing values.
        verbose (bool):
            Whether to print additional information.

    Returns:
        Tuple[pd.DataFrame, pd.Series]:
            A tuple containing the forward and backward filled DataFrame and a numeric series (0.0 or 1.0) where 0.0 indicates a weight for missing values/gaps.

    Examples:
        >>> from spotforecast2_safe.data.fetch_data import fetch_data
        >>> from spotforecast2_safe.preprocessing.imputation import get_missing_weights
        >>> data = fetch_data()
        >>> filled_data, missing_weights = get_missing_weights(data, window_size=72, verbose=True)

    """
    # first perform some checks if dataframe has enough data and if window_size is appropriate
    if data.shape[0] == 0:
        raise ValueError("Input data is empty.")
    if window_size <= 0:
        raise ValueError("window_size must be a positive integer.")
    if window_size >= data.shape[0]:
        raise ValueError("window_size must be smaller than the number of rows in data.")

    missing_indices = data.index[data.isnull().any(axis=1)]
    n_missing = len(missing_indices)
    if verbose:
        pct_missing = (n_missing / len(data)) * 100
        print(f"Number of rows with missing values: {n_missing}")
        print(f"Percentage of rows with missing values: {pct_missing:.2f}%")
        print(f"missing_indices: {missing_indices}")
    data = data.ffill()
    data = data.bfill()

    is_missing = pd.Series(0, index=data.index)
    is_missing.loc[missing_indices] = 1
    # weights_series: 0 if in/near gap, 1 otherwise
    weights_series = (
        1.0 - is_missing.rolling(window=window_size + 1, min_periods=1).max()
    )

    if verbose:
        n_missing_after = (weights_series == 0).sum()
        pct_missing_after = (n_missing_after / len(data)) * 100
        print(
            f"Number of rows with missing weights after processing: {n_missing_after}"
        )
        print(
            f"Percentage of rows with missing weights after processing: {pct_missing_after:.2f}%"
        )
    return data, weights_series



# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Module for managing model training.
"""

import glob
import logging
import re
from pathlib import Path
from typing import Any, Optional, Union

from joblib import load

from spotforecast2_safe.data.fetch_data import get_cache_home

logger = logging.getLogger(__name__)


def get_path_model(
    name: str,
    iteration: int,
    model_dir: Optional[Union[str, Path]] = None,
) -> Path:
    """Yield the path to a model file for a given iteration and model name.

    Args:
        name: Model name (e.g. ``"lgbm"``, ``"xgb"``).
        iteration: Iteration of the model.
        model_dir: Directory where models are stored.
            If *None*, defaults to :func:`get_cache_home`.

    Returns:
        Path: Full path where the model file should be stored.

    Examples:
        >>> import tempfile
        >>> from pathlib import Path
        >>> from spotforecast2_safe.manager.trainer import get_path_model
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     p = get_path_model("lgbm", 3, model_dir=tmpdir)
        ...     p.name
        'lgbm_forecaster_3.joblib'
    """
    if model_dir is None:
        model_dir = get_cache_home()
    else:
        model_dir = Path(model_dir)
    return model_dir / f"{name}_forecaster_{iteration}.joblib"


def load_iteration(
    name: str,
    iteration: int,
    model_dir: Optional[Union[str, Path]] = None,
) -> Optional[Any]:
    """Load a saved model at a given iteration.

    Args:
        name: Model name (e.g. ``"lgbm"``).
        iteration: Iteration of the model.
        model_dir: Directory where models are stored.
            If *None*, defaults to :func:`get_cache_home`.

    Returns:
        The loaded model instance, or *None* if the file does not exist.

    Examples:
        >>> import tempfile
        >>> from spotforecast2_safe.manager.trainer import load_iteration
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     result = load_iteration("lgbm", 99, model_dir=tmpdir)
        ...     result is None
        True
    """
    path_file = get_path_model(name, iteration, model_dir=model_dir)
    if not path_file.exists():
        logger.error("Iteration %d does not exist at %s!", iteration, path_file)
        return None
    try:
        model = load(path_file)
        return model
    except Exception as e:
        logger.error("Failed to load model from %s: %s", path_file, e)
        return None


def get_last_model(
    model_name: str, model_dir: Optional[Union[str, Path]] = None
) -> tuple[int, Any]:
    """
    Get the latest trained model from the cache.

    Args:
        model_name: Name of the model (e.g., 'lgbm', 'xgb').
        model_dir: Directory where models are stored. If None, defaults to
            the library's cache home.

    Returns:
        A tuple (iteration, model_instance). If no model is found,
        returns (-1, None).
    """
    if model_dir is None:
        model_dir = get_cache_home()
    else:
        model_dir = Path(model_dir)

    if not model_dir.exists():
        return -1, None

    list_files = glob.glob(str(model_dir / f"{model_name}_forecaster_*.joblib"))
    if not list_files:
        return -1, None

    searches = [
        re.search(rf"{model_name}_forecaster_(\d+)\.joblib", x) for x in list_files
    ]
    iterations = [int(search.group(1)) for search in searches if search is not None]

    if not iterations:
        return -1, None

    max_iter = max(iterations)
    file_path = model_dir / f"{model_name}_forecaster_{max_iter}.joblib"

    try:
        model = load(file_path)
        return max_iter, model
    except Exception as e:
        logger.error("Failed to load model from %s: %s", file_path, e)
        return -1, None



# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later

"""Data structures for input and processed data."""

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Tuple

import pandas as pd

from spotforecast2_safe.utils.convert_to_utc import convert_to_utc

logger = logging.getLogger(__name__)


@dataclass
class Data:
    """Container for input time series data.

    Attributes:
        data: pandas DataFrame containing the input time series data.
    """

    data: pd.DataFrame

    @classmethod
    def from_csv(
        cls,
        csv_path: Path,
        timezone: Optional[str],
        columns: Optional[List[str]] = None,
        parse_dates=True,
        index_col=0,
        **kwargs,
    ) -> "Data":
        """Load data from a CSV file.

        The CSV must contain a datetime column that becomes the DataFrame index.
        The index is localized to the provided timezone if it is naive, and then
        converted to UTC.

        Args:
            csv_path (Path): Path to the CSV file.
            timezone (Optional[str]): Timezone to assign if the index has no
                timezone. Must be provided if the index is naive.
            columns (Optional[List[str]]): List of column names to include. If
                provided, only these columns will be loaded from the CSV
                (optimizes reading speed). If None, all columns are loaded.
            parse_dates (bool or list, optional): Passed to ``pd.read_csv``.
                Defaults to True.
            index_col (int or str, optional): Column to use as index. Defaults to 0.
            **kwargs (Any): Additional keyword arguments forwarded to ``pd.read_csv``.

        Returns:
            Data: Instance containing the loaded DataFrame.

        Raises:
            ValueError: If the CSV does not yield a DatetimeIndex.
            ValueError: If the index is timezone-naive and no timezone is provided.

        Examples:
            >>> from spotforecast2_safe.data import Data
            >>> data = Data.from_csv(
            ...     Path("data.csv"),
            ...     timezone="UTC",
            ...     columns=["target_col"]
            ... )
        """
        # If columns specified, add index column to usecols for efficient reading
        usecols = None
        if columns is not None:
            # Get the index column name/number
            if isinstance(index_col, int):
                # Read header first to get column names
                header_df = pd.read_csv(csv_path, nrows=0)
                index_col_name = header_df.columns[index_col]
            else:
                index_col_name = index_col
            usecols = [index_col_name] + columns

        df = pd.read_csv(
            csv_path,
            parse_dates=parse_dates,
            index_col=index_col,
            usecols=usecols,
            **kwargs,
        )
        df = convert_to_utc(df, timezone)
        if df.index.freq is None:
            try:
                df.index.freq = pd.infer_freq(df.index)
            except (ValueError, TypeError) as e:
                logger.debug(
                    "Could not infer frequency for index in from_csv: %s. "
                    "Frequency will remain None.",
                    e,
                )
        return cls(data=df)

    @classmethod
    def from_dataframe(
        cls,
        df: pd.DataFrame,
        timezone: Optional[str],
        columns: Optional[List[str]] = None,
    ) -> "Data":
        """Create a new Data instance from an existing DataFrame.

        The DataFrame must have a datetime index. The index is localized to the
        provided timezone if it is naive, and then converted to UTC.

        Args:
            df (pd.DataFrame): Input DataFrame containing data.
            timezone (Optional[str]): Timezone to assign if the index is naive.
                Must be provided if the index has no timezone.
            columns (Optional[List[str]]): List of column names to include.
                If provided, only these columns will be selected from the
                DataFrame. If None, all columns are used.

        Returns:
            Data: Instance containing the provided DataFrame.

        Raises:
            ValueError: If the DataFrame index is not a DatetimeIndex.
            ValueError: If the index is timezone-naive and no timezone is provided.
        """
        df = convert_to_utc(df, timezone)

        # Select columns if specified
        if columns is not None:
            df = df[columns].copy()

        if df.index.freq is None:
            try:
                df.index.freq = pd.infer_freq(df.index)
            except (ValueError, TypeError) as e:
                logger.debug(
                    "Could not infer frequency for index in from_dataframe: %s. "
                    "Frequency will remain None.",
                    e,
                )

        return cls(data=df)


@dataclass(frozen=True)
class Period:
    """Class abstraction for the information required to encode a period.

    Attributes:
        name: Name of the period (e.g., 'hour', 'day').
        n_periods: Number of periods to encode (e.g., 24 for hours).
        column: Name of the column in the DataFrame containing the period information.
        input_range: Tuple of (min, max) values for the period (e.g., (0, 23) for hours).

    Examples:
        >>> from spotforecast2_safe.data import Period
        >>> period = Period(name="hour", n_periods=24, column="hour", input_range=(0, 23))
        >>> period.name
        'hour'
    """

    name: str
    n_periods: int
    column: str
    input_range: Tuple[int, int]

    def __post_init__(self):
        """Validate the period configuration."""
        if self.n_periods <= 0:
            raise ValueError(f"n_periods must be positive, got {self.n_periods}")
        if self.input_range[0] >= self.input_range[1]:
            raise ValueError(
                f"input_range[0] must be less than input_range[1], got {self.input_range}"
            )



# SPDX-FileCopyrightText: skforecast team
# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later AND BSD-3-Clause

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted
from ._common import _check_X_numpy_ndarray_1d


class TimeSeriesDifferentiator(BaseEstimator, TransformerMixin):
    """
    Transforms a time series into a differenced time series.

    Args:
        order (int, optional): Order of differentiation. Defaults to 1.
        window_size (int, optional): The window size used by the forecaster. Defaults to None.
        initial_values (list, numpy ndarray, optional): Values to be used for the inverse transformation (reverting differentiation).
            If None, the first `order` values of the training data `X` are stored during `fit`.

    Attributes:
        initial_values_ (list): Values stored for inverse transformation.
        last_values_ (list): Last values of the differenced time series.
        pre_train_values_ (list): First training values for inverse transformation of training data.
    """

    def __init__(
        self,
        order: int = 1,
        window_size: int | None = None,
        initial_values: list | np.ndarray | None = None,
    ):
        self.order = order
        self.window_size = window_size
        self.initial_values = initial_values

    @_check_X_numpy_ndarray_1d(ensure_1d=True)
    def fit(self, X: np.ndarray, y: object = None) -> object:
        """
        Store initial values if not provided.
        """
        if self.order < 1:
            raise ValueError("`order` must be a positive integer.")

        if self.window_size is not None:
            if not isinstance(self.window_size, int):
                raise TypeError(
                    f"Parameter `window_size` must be an integer greater than 0. "
                    f"Found {type(self.window_size)}."
                )
            if self.window_size < 1:
                raise ValueError(
                    f"Parameter `window_size` must be an integer greater than 0. "
                    f"Found {self.window_size}."
                )

        self.initial_values_ = []
        self.pre_train_values_ = []
        self.last_values_ = []

        if self.initial_values is None:
            if len(X) < self.order:
                raise ValueError(
                    f"The time series must have at least {self.order} values "
                    f"to compute the differentiation of order {self.order}."
                )

            # Logic similar to Skforecast:
            # We iterate to capture initial values, pre_train_values, and last_values
            # Skforecast does this iteratively. The original spotforecast implementation
            # was simplified. Integrating window_size support requires the iterative approach
            # or careful indexing.

            # Re-implementing skforecast logic for robustness with window_size

            current_X = X.copy()
            for i in range(self.order):
                self.initial_values_.append(current_X[0])
                if self.window_size is not None:
                    # Skforecast logic: self.pre_train_values.append(X[self.window_size - self.order])
                    # But wait, skforecast loop updates X_diff. A recursive implementation.

                    # If we follow skforecast exactly:
                    # if i == 0: X_diff = diff(X); initial = X[0]; pre = X[ws-order]; last = X[-1]
                    # else: X_diff = diff(previous_diff); initial = prev_diff[0]; pre = prev_diff[ws-order]; last = prev_diff[-1]

                    # Current X is the series being differentiated in this step
                    if len(current_X) > (self.window_size - self.order):
                        self.pre_train_values_.append(
                            current_X[self.window_size - self.order]
                        )
                    else:
                        # Fallback if X is smaller than window_size (shouldn't happen during training if window_size is correct)
                        self.pre_train_values_.append(np.nan)

                self.last_values_.append(current_X[-1])
                current_X = np.diff(current_X, n=1)

        else:
            if len(self.initial_values) != self.order:
                raise ValueError(
                    f"The length of `initial_values` must be equal to the order "
                    f"of differentiation ({self.order})."
                )
            self.initial_values_ = list(self.initial_values)
            # If initial_values provided, we can still compute last_values_ from X if we assume X is the training data
            # But usually initial_values are provided when loading/restoring.
            # For now, let's keep the user provided initial values.
            # But we still need last_values_ for next window.
            # Assuming X is the training data passed to fit()

            self.last_values_ = list(X[-self.order :])

        return self

    @_check_X_numpy_ndarray_1d(ensure_1d=True)
    def fit_transform(self, X: np.ndarray, y: object = None) -> np.ndarray:
        """
        Fit and transform.
        """
        return self.fit(X).transform(X)

    @_check_X_numpy_ndarray_1d(ensure_1d=True)
    def transform(self, X: np.ndarray, y: object = None) -> np.ndarray:
        """
        Compute the differences.
        """
        if not hasattr(self, "initial_values_") and self.initial_values is not None:
            self.fit(X)
        elif not hasattr(self, "initial_values_"):
            check_is_fitted(self, ["initial_values_"])

        X_diff = np.diff(X, n=self.order)
        # Pad with NaNs to keep same length
        X_diff = np.concatenate([np.full(self.order, np.nan), X_diff])

        # Update last values seen (for next window inverse)
        self.last_values_ = X[-self.order :]

        return X_diff

    def inverse_transform_next_window(self, X: np.ndarray) -> np.ndarray:
        """
        Inverse transform for the next window of predictions.
        """
        check_is_fitted(self, ["initial_values_", "last_values_"])

        if self.order == 1:
            result = np.cumsum(X) + self.last_values_[-1]
        else:
            # Recursive or iterative approach for higher orders
            # Simplified: Assuming order 1 is sufficient for now or throwing error
            raise NotImplementedError(
                "inverse_transform_next_window not implemented for order > 1"
            )

        return result

    @_check_X_numpy_ndarray_1d(ensure_1d=True)
    def inverse_transform(self, X: np.ndarray, y: object = None) -> np.ndarray:
        """
        Revert the differences.
        """
        check_is_fitted(self, ["initial_values_"])

        # X contains the differenced series (with NaNs at the beginning potentially)
        # remove NaNs at the start corresponding to order
        X_clean = X[self.order :]

        if len(X_clean) == 0:
            # Just return initial values if only NaNs were passed
            return np.array(self.initial_values_)

        result = list(self.initial_values_)

        if self.order == 1:
            current_value = result[-1]
            restored = []
            for diff_val in X_clean:
                current_value += diff_val
                restored.append(current_value)
            result.extend(restored)
        else:
            # Recursive reconstruction for higher orders logic check
            # For order > 1, np.diff does repeated diffs.
            # To invert, we need to do repeated cumsum.
            # But we need appropriate initial values for each level of integration.
            # This is a simplified version.

            raise NotImplementedError(
                "Inverse transform for order > 1 is currently not fully implemented in this port."
            )

        return np.array(result)

    @_check_X_numpy_ndarray_1d(ensure_1d=True)
    def inverse_transform_training(self, X: np.ndarray, y: object = None) -> np.ndarray:
        """
        Reverts the differentiation for training data.
        """
        if not hasattr(self, "pre_train_values_") or not self.pre_train_values_:
            raise ValueError(
                "The `window_size` parameter must be set before fitting the "
                "transformer to revert the differentiation of the training "
                "time series."
            )

        # Implementation of inverse_transform_training logic
        # For order 1
        if self.order == 1:
            X_clean = X[self.order :]
            # Reconstruct
            # pre_train_values_ contains the value right BEFORE the training window starts?
            # skforecast: X_undiff = np.insert(X, 0, self.pre_train_values[-1])
            #             X_undiff = np.cumsum(X_undiff)
            #             X_undiff = X_undiff[self.order:]

            X_undiff = np.insert(X_clean, 0, self.pre_train_values_[-1])
            X_undiff = np.cumsum(X_undiff)
            # The skforecast logic seems to insert, cumsum, then slice.
            # If X_clean is the differentiated training data (which generates y_train),
            # we need the value just before it to start cumsum.

            # Simplified for order=1 for now to match safety/robustness needs.
            return X_undiff

        else:
            raise NotImplementedError(
                "inverse_transform_training not implemented for order > 1"
            )



# SPDX-FileCopyrightText: skforecast team
# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later AND BSD-3-Clause


"""ForecasterBase class.

This module contains the base class for all forecasters in spotforecast2_safe and spotforecast.
All forecasters should specify all the parameters that can be set at the
class level in their __init__.

Examples:
    Create a custom forecaster inheriting from ForecasterBase:

    >>> from spotforecast2_safe.forecaster.base import ForecasterBase
    >>> import pandas as pd
    >>> import numpy as np
    >>> class MyForecaster(ForecasterBase):
    ...     def __init__(self, estimator):
    ...         self.estimator = estimator
    ...         self.__spotforecast_tags__ = {'hide_lags': True}
    ...     def create_train_X_y(self, y, exog=None):
    ...         return pd.DataFrame(), pd.Series(dtype=float)
    ...     def fit(self, y, exog=None):
    ...         pass
    ...     def predict(self, steps, last_window=None, exog=None):
    ...         return pd.Series(np.zeros(steps))
    ...     def set_params(self, params):
    ...         pass
    >>> from sklearn.linear_model import Ridge
    >>> forecaster = MyForecaster(estimator=Ridge())
    >>> forecaster
    MyForecaster(estimator=Ridge())
"""

from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any
import textwrap
import warnings
import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline


class ForecasterBase(ABC):
    """Base class for all forecasters in spotforecast2.

    All forecasters should specify all the parameters that can be set at
    the class level in their __init__.

    Attributes:
        __spotforecast_tags__: Dictionary with forecaster tags that characterize
            the behavior of the forecaster.

    Examples:
        To see all abstract methods that need to be implemented:

        >>> import inspect
        >>> from spotforecast2_safe.forecaster.base import ForecasterBase
        >>> [m[0] for m in inspect.getmembers(ForecasterBase, predicate=inspect.isabstract)]
        ['create_train_X_y', 'fit', 'predict', 'set_params']
    """

    def _preprocess_repr(
        self,
        estimator: object | None = None,
        training_range_: dict[str, str] | None = None,
        series_names_in_: list[str] | None = None,
        exog_names_in_: list[str] | None = None,
        transformer_series: object | dict[str, object] | None = None,
    ) -> tuple[str, str | None, str | None, str | None, str | None]:
        """Prepare the information to be displayed when a Forecaster object is printed.

        Args:
            estimator: Estimator object. Default is None.
            training_range_: Training range. Only used for ForecasterRecursiveMultiSeries.
                Default is None.
            series_names_in_: Names of the series used in the forecaster.
                Only used for ForecasterRecursiveMultiSeries. Default is None.
            exog_names_in_: Names of the exogenous variables used in the forecaster.
                Default is None.
            transformer_series: Transformer used in the series.
                Only used for ForecasterRecursiveMultiSeries. Default is None.

        Returns:
            Tuple containing params (estimator parameters string), training_range_
            (training range string representation), series_names_in_ (series names
            string representation), exog_names_in_ (exogenous variable names string
            representation), and transformer_series (transformer string representation).

        Examples:
            >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
            >>> from sklearn.linear_model import Ridge
            >>> estimator = Ridge(alpha=0.5)
            >>> forecaster = ForecasterRecursive(estimator=estimator, lags=3)
            >>> params, tr, sn, en, ts = forecaster._preprocess_repr(estimator=estimator)
            >>> params
            "{'alpha': 0.5, 'copy_X': True, 'fit_intercept': True, 'max_iter': None, 'positive': False, 'random_state': None, 'solver': 'auto', 'tol': 0.0001}"
        """

        if estimator is not None:
            if isinstance(estimator, Pipeline):
                name_pipe_steps = tuple(
                    name + "__" for name in estimator.named_steps.keys()
                )
                params = {
                    key: value
                    for key, value in estimator.get_params().items()
                    if key.startswith(name_pipe_steps)
                }
            else:
                params = estimator.get_params()
            params = str(params)
        else:
            params = None

        if training_range_ is not None:
            training_range_ = [
                f"'{k}': {v.astype(str).to_list()}" for k, v in training_range_.items()
            ]
            if len(training_range_) > 10:
                training_range_ = training_range_[:5] + ["..."] + training_range_[-5:]
            training_range_ = ", ".join(training_range_)

        if series_names_in_ is not None:
            if len(series_names_in_) > 50:
                series_names_in_ = (
                    series_names_in_[:25] + ["..."] + series_names_in_[-25:]
                )
            series_names_in_ = ", ".join(series_names_in_)

        if exog_names_in_ is not None:
            if len(exog_names_in_) > 50:
                exog_names_in_ = exog_names_in_[:25] + ["..."] + exog_names_in_[-25:]
            exog_names_in_ = ", ".join(exog_names_in_)

        if transformer_series is not None:
            if isinstance(transformer_series, dict):
                transformer_series = [
                    f"'{k}': {v}" for k, v in transformer_series.items()
                ]
                if len(transformer_series) > 10:
                    transformer_series = (
                        transformer_series[:5] + ["..."] + transformer_series[-5:]
                    )
                transformer_series = ", ".join(transformer_series)
            else:
                transformer_series = str(transformer_series)

        return (
            params,
            training_range_,
            series_names_in_,
            exog_names_in_,
            transformer_series,
        )

    def _format_text_repr(
        self,
        text: str,
        max_text_length: int = 58,
        width: int = 80,
        indent: str = "    ",
    ) -> str:
        """Format text for __repr__ method.

        Args:
            text: Text to format.
            max_text_length: Maximum length of the text before wrapping. Default is 58.
            width: Maximum width of the text. Default is 80.
            indent: Indentation of the text. Default is four spaces.

        Returns:
            Formatted text string with proper wrapping and indentation.

        Examples:
            >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
            >>> from sklearn.linear_model import Ridge
            >>> forecaster = ForecasterRecursive(estimator=Ridge(), lags=3)
            >>> forecaster._format_text_repr("Short text")
            'Short text'
        """

        if text is not None and len(text) > max_text_length:
            text = "\n    " + textwrap.fill(
                str(text), width=width, subsequent_indent=indent
            )

        return text

    @abstractmethod
    def create_train_X_y(
        self, y: pd.Series, exog: pd.Series | pd.DataFrame | None = None
    ) -> tuple[pd.DataFrame, pd.Series]:
        """Create training matrices from univariate time series and exogenous variables.

        Args:
            y: Training time series.
            exog: Exogenous variable(s) included as predictor(s). Must have the same
                number of observations as y and their indexes must be aligned.
                Default is None.

        Returns:
            Tuple containing X_train (training values/predictors with shape
            (len(y) - max_lag, len(lags))) and y_train (target values of the
            time series related to each row of X_train with shape (len(y) - max_lag,)).

        Examples:
            >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
            >>> from sklearn.linear_model import Ridge
            >>> import pandas as pd
            >>> import numpy as np
            >>> forecaster = ForecasterRecursive(estimator=Ridge(), lags=3)
            >>> y = pd.Series(np.arange(10), name='y')
            >>> X_train, y_train = forecaster.create_train_X_y(y)
            >>> X_train.head(2)
               lag_1  lag_2  lag_3
            3    2.0    1.0    0.0
            4    3.0    2.0    1.0
            >>> y_train.head(2)
            3    3
            4    4
            Name: y, dtype: int64
        """

        pass

    @abstractmethod
    def fit(self, y: pd.Series, exog: pd.Series | pd.DataFrame | None = None) -> None:
        """Training Forecaster.

        Args:
            y: Training time series.
            exog: Exogenous variable(s) included as predictor(s). Must have the same
                number of observations as y and their indexes must be aligned so
                that y[i] is regressed on exog[i]. Default is None.

        Returns:
            None

        Examples:
            >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
            >>> from sklearn.linear_model import Ridge
            >>> import pandas as pd
            >>> import numpy as np
            >>> forecaster = ForecasterRecursive(estimator=Ridge(), lags=3)
            >>> y = pd.Series(np.arange(10), name='y')
            >>> forecaster.fit(y)
            >>> forecaster.is_fitted
            True
        """

        pass

    @abstractmethod
    def predict(
        self,
        steps: int,
        last_window: pd.Series | pd.DataFrame | None = None,
        exog: pd.Series | pd.DataFrame | None = None,
    ) -> pd.Series:
        """Predict n steps ahead.

        Args:
            steps: Number of steps to predict.
            last_window: Series values used to create the predictors (lags) needed in the
                first iteration of the prediction (t + 1). If None, the values stored in
                last_window are used to calculate the initial predictors, and the
                predictions start right after training data. Default is None.
            exog: Exogenous variable(s) included as predictor(s). Default is None.

        Returns:
            Predicted values as a pandas Series.

        Examples:
            >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
            >>> from sklearn.linear_model import Ridge
            >>> import pandas as pd
            >>> import numpy as np
            >>> forecaster = ForecasterRecursive(estimator=Ridge(), lags=3)
            >>> y = pd.Series(np.arange(10), name='y')
            >>> forecaster.fit(y)
            >>> forecaster.predict(steps=3)
            10    9.5
            11    9.0
            12    8.5
            Name: pred, dtype: float64
        """

        pass

    @abstractmethod
    def set_params(self, params: dict[str, object]) -> None:
        """Set new values to the parameters of the scikit-learn model stored in the forecaster.

        Args:
            params: Parameters values dictionary.

        Returns:
            None

        Examples:
            >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
            >>> from sklearn.linear_model import Ridge
            >>> forecaster = ForecasterRecursive(estimator=Ridge(alpha=1.0), lags=3)
            >>> forecaster.set_params({'estimator__alpha': 0.5})
            >>> forecaster.estimator.alpha
            0.5
        """

        pass

    def set_lags(
        self, lags: int | list[int] | np.ndarray[int] | range[int] | None = None
    ) -> None:
        """Set new value to the attribute lags.

        Attributes max_lag and window_size are also updated.

        Args:
            lags: Lags used as predictors. Index starts at 1, so lag 1 is equal to t-1.
                If int: include lags from 1 to lags (included). If list, 1d numpy ndarray,
                or range: include only lags present in lags, all elements must be int.
                If None: no lags are included as predictors. Default is None.

        Returns:
            None

        Examples:
            >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
            >>> from sklearn.linear_model import Ridge
            >>> forecaster = ForecasterRecursive(estimator=Ridge(), lags=3)
            >>> forecaster.set_lags(lags=5)
            >>> forecaster.lags
            array([1, 2, 3, 4, 5])
        """

        pass

    def set_window_features(
        self, window_features: object | list[object] | None = None
    ) -> None:
        """Set new value to the attribute window_features.

        Attributes max_size_window_features, window_features_names,
        window_features_class_names and window_size are also updated.

        Args:
            window_features: Instance or list of instances used to create window features.
                Window features are created from the original time series and are
                included as predictors. Default is None.

        Returns:
            None

        Examples:
            >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
            >>> from spotforecast2_safe.forecaster.preprocessing import RollingFeatures
            >>> from sklearn.linear_model import Ridge
            >>> forecaster = ForecasterRecursive(estimator=Ridge(), lags=3)
            >>> window_feat = RollingFeatures(stats='mean', window_sizes=3)
            >>> forecaster.set_window_features(window_features=window_feat)
            >>> forecaster.window_features
            [RollingFeatures(stats=['mean'], window_sizes=[3])]
        """

        pass

    def get_tags(self) -> dict[str, Any]:
        """Return the tags that characterize the behavior of the forecaster.

        Returns:
            Dictionary with forecaster tags describing behavior and capabilities.

        Examples:
            >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
            >>> from sklearn.linear_model import Ridge
            >>> forecaster = ForecasterRecursive(estimator=Ridge(), lags=3)
            >>> tags = forecaster.get_tags()
            >>> tags['forecaster_task']
            'regression'
        """

        return self.__spotforecast_tags__

    def summary(self) -> None:
        """Show forecaster information.

        Returns:
            None

        Examples:
            >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
            >>> from sklearn.linear_model import Ridge
            >>> forecaster = ForecasterRecursive(estimator=Ridge(), lags=3)
            >>> forecaster.summary()
            ForecasterRecursive
            ===================
            Estimator: Ridge()
            Lags: [1 2 3]
            ...
        """

        print(self.__repr__())

    def __setstate__(self, state: dict) -> None:
        """Custom __setstate__ to ensure backward compatibility when unpickling.

        This method is called when an object is unpickled (deserialized).
        It handles the migration of deprecated attributes to their new names.

        Args:
            state: The state dictionary from the pickled object.

        Returns:
            None

        Examples:
            >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
            >>> from sklearn.linear_model import Ridge
            >>> import pickle
            >>> forecaster = ForecasterRecursive(estimator=Ridge(), lags=3)
            >>> pickled_forecaster = pickle.dumps(forecaster)
            >>> unpickled_forecaster = pickle.loads(pickled_forecaster)
        """

        if "regressor" in state and "estimator" not in state:
            state["estimator"] = state.pop("regressor")

        self.__dict__.update(state)

    @property
    def regressor(self) -> Any:
        """Deprecated property. Use estimator instead.

        Returns:
            The estimator object.

        Examples:
            >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
            >>> from sklearn.linear_model import Ridge
            >>> forecaster = ForecasterRecursive(estimator=Ridge(), lags=3)
            >>> forecaster.regressor # Raises FutureWarning
            Ridge()
        """
        warnings.warn(
            "The `regressor` attribute is deprecated and will be removed in future "
            "versions. Use `estimator` instead.",
            FutureWarning,
        )
        return self.estimator



# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later

"""Exogenous feature builder for time series forecasting."""

from typing import List, Optional

import holidays
import pandas as pd

from spotforecast2_safe.data.data import Period
from spotforecast2_safe.preprocessing.repeating_basis_function import (
    RepeatingBasisFunction,
)


class ExogBuilder:
    """
    Builds a set of exogenous features for a given date range.

    This builder combines temporal features (day of year, day of week, hour, etc.)
    with cyclical features encoded via RepeatingBasisFunctions and optional
    holiday indicators.

    Attributes:
        periods (List[Period]): List of periodic features to encode.
        country_code (Optional[str]): Country code for holiday lookups.
        holidays_list (Optional[holidays.HolidayBase]): List of holidays for the specified country.

    Examples:
        >>> import pandas as pd
        >>> from spotforecast2_safe.data.data import Period
        >>> from spotforecast2_safe.preprocessing.exog_builder import ExogBuilder
        >>> periods = [Period(name="hour", n_periods=24, column="hour", input_range=(0, 23))]
        >>> builder = ExogBuilder(periods=periods, country_code="DE")
        >>> start = pd.Timestamp("2025-01-01", tz="UTC")
        >>> end = pd.Timestamp("2025-01-02", tz="UTC")
        >>> exog = builder.build(start, end)
        >>> exog.shape[1] > 0
        True
    """

    def __init__(
        self, periods: Optional[List[Period]] = None, country_code: Optional[str] = None
    ):
        """
        Initialize the ExogBuilder.

        Args:
            periods: List of Period objects defining cyclical features.
            country_code: country code (ISO) for holiday detection.
        """
        self.periods = periods or []
        self.country_code = country_code
        self.holidays_list = (
            holidays.country_holidays(country_code) if country_code else None
        )

    def _get_time_columns(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Extract basic time-based columns from the DataFrame index.

        Args:
            X: DataFrame with DatetimeIndex.

        Returns:
            pd.DataFrame: Copy of X with extra time columns.
        """
        X = X.copy()
        X["dayofyear"] = X.index.dayofyear
        X["dayofweek"] = X.index.dayofweek
        X["quarter"] = X.index.quarter
        X["month"] = X.index.month
        X["hour"] = X.index.hour
        return X

    def build(self, start_date: pd.Timestamp, end_date: pd.Timestamp) -> pd.DataFrame:
        """
        Build the exogenous feature DataFrame for a date range.

        The generated DataFrame has an hourly frequency.

        Args:
            start_date: Start of the date range (inclusive).
            end_date: End of the date range (inclusive).

        Returns:
            pd.DataFrame: DataFrame containing exogenous features.

        Raises:
            ValueError: If the date range is invalid.

        Examples:
            >>> import pandas as pd
            >>> from spotforecast2_safe.data.data import Period
            >>> from spotforecast2_safe.preprocessing.exog_builder import ExogBuilder
            >>> periods = [Period(name="hour", n_periods=24, column="hour", input_range=(0, 23))]
            >>> builder = ExogBuilder(periods=periods, country_code="DE")
            >>> start = pd.Timestamp("2025-01-01", tz="UTC")
            >>> end = pd.Timestamp("2025-01-02", tz="UTC")
            >>> exog = builder.build(start, end)
            >>> exog.shape[1] > 0
            True
        """
        date_range = pd.date_range(start=start_date, end=end_date, freq="h")
        X = pd.DataFrame(index=date_range)
        X = self._get_time_columns(X)

        seasons_encoded = []
        for period in self.periods:
            rbf = RepeatingBasisFunction(
                n_periods=period.n_periods,
                column=period.column,
                input_range=period.input_range,
            )
            season_encoded = rbf.transform(X)
            cols = [f"{period.name}_{i}" for i in range(season_encoded.shape[1])]
            seasons_encoded.append(
                pd.DataFrame(season_encoded, index=X.index, columns=cols)
            )

        X_ = pd.concat(seasons_encoded, axis=1) if seasons_encoded else X

        if self.holidays_list is not None:
            # List comprehension is robust for holiday detection across different
            # pandas/holidays versions and handling of DatetimeIndex
            X_["holidays"] = [int(d in self.holidays_list) for d in X_.index]

        X_["is_weekend"] = X_.index.dayofweek.isin([5, 6]).astype(int)
        return X_



# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later

from .base import ForecasterBase
from .recursive import ForecasterRecursive

__all__ = [
    "ForecasterBase",
    "ForecasterRecursive",
]



# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later

from spotforecast2_safe.data.data import Data, Period
from spotforecast2_safe.data.fetch_data import (
    fetch_data,
    fetch_holiday_data,
    fetch_weather_data,
    get_cache_home,
    get_data_home,
)

__all__ = [
    "Data",
    "Period",
    "get_data_home",
    "get_cache_home",
    "fetch_data",
    "fetch_holiday_data",
    "fetch_weather_data",
]



# SPDX-FileCopyrightText: skforecast team
# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later AND BSD-3-Clause

from __future__ import annotations
from typing import Callable
from copy import deepcopy
import warnings
import numpy as np
import pandas as pd
from joblib import Parallel, delayed, cpu_count
from tqdm.auto import tqdm

from spotforecast2_safe.forecaster.metrics import add_y_train_argument, _get_metric
from spotforecast2_safe.exceptions import (
    LongTrainingWarning,
    IgnoredArgumentWarning,
    set_skforecast_warnings,
)
from spotforecast2_safe.model_selection.split_ts_cv import TimeSeriesFold
from spotforecast2_safe.model_selection.split_one_step import OneStepAheadFold
from spotforecast2_safe.model_selection.utils_common import (
    check_backtesting_input,
    check_one_step_ahead_input,
    select_n_jobs_backtesting,
)


def _backtesting_forecaster(
    forecaster: object,
    y: pd.Series,
    cv: TimeSeriesFold,
    metric: str | Callable | list[str | Callable],
    exog: pd.Series | pd.DataFrame | None = None,
    interval: float | list[float] | tuple[float] | str | object | None = None,
    interval_method: str = "bootstrapping",
    n_boot: int = 250,
    use_in_sample_residuals: bool = True,
    use_binned_residuals: bool = True,
    random_state: int = 123,
    return_predictors: bool = False,
    n_jobs: int | str = "auto",
    verbose: bool = False,
    show_progress: bool = True,
    suppress_warnings: bool = False,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Backtesting of forecaster model following the folds generated by the TimeSeriesFold
    class and using the metric(s) provided.

    If `forecaster` is already trained and `initial_train_size` is set to `None` in the
    TimeSeriesFold class, no initial train will be done and all data will be used
    to evaluate the model. However, the first `len(forecaster.last_window)` observations
    are needed to create the initial predictors, so no predictions are calculated for
    them.

    A copy of the original forecaster is created so that it is not modified during the
    process.

    Args:
        forecaster (ForecasterRecursive, ForecasterDirect, ForecasterEquivalentDate):
            Forecaster model.
        y (pd.Series): Training time series.
        cv (TimeSeriesFold): TimeSeriesFold object with the information needed to
            split the data into folds.
        metric (str | Callable | list): Metric used to quantify the goodness of fit
            of the model.

            - If `str`: {'mean_squared_error', 'mean_absolute_error',
              'mean_absolute_percentage_error', 'mean_squared_log_error',
              'mean_absolute_scaled_error', 'root_mean_squared_scaled_error'}
            - If `Callable`: Function with arguments `y_true`, `y_pred` and `y_train`
              (Optional) that returns a float.
            - If `list`: List containing multiple strings and/or Callables.
        exog (pd.Series | pd.DataFrame, optional): Exogenous variable/s included as
            predictor/s. Must have the same number of observations as `y` and should
            be aligned so that y[i] is regressed on exog[i]. Defaults to None.
        interval (float | list | tuple | str | object, optional): Specifies whether
            probabilistic predictions should be estimated and the method to use.
            The following options are supported:

            - If `float`, represents the nominal (expected) coverage (between 0 and 1).
            For instance, `interval=0.95` corresponds to `[2.5, 97.5]` percentiles.
            - If `list` or `tuple`: Sequence of percentiles to compute, each value must
              be between 0 and 100 inclusive. For example, a 95% confidence interval can
              be specified as `interval = [2.5, 97.5]` or multiple percentiles (e.g. 10,
              50 and 90) as `interval = [10, 50, 90]`.
            - If 'bootstrapping' (str): `n_boot` bootstrapping predictions will be
              generated.
            - If scipy.stats distribution object, the distribution parameters will
              be estimated for each prediction.
            - If None, no probabilistic predictions are estimated.
            Defaults to None.
        interval_method (str, optional): Technique used to estimate prediction
            intervals. Available options:

            - 'bootstrapping': Bootstrapping is used to generate prediction intervals.
            - 'conformal': Employs the conformal prediction split method for
              interval estimation.
            Defaults to 'bootstrapping'.
        n_boot (int, optional): Number of bootstrapping iterations to perform when
            estimating prediction intervals. Defaults to 250.
        use_in_sample_residuals (bool, optional): If `True`, residuals from the
            training data are used as proxy of prediction error to create predictions.
            If `False`, out of sample residuals (calibration) are used.
            Out-of-sample residuals must be precomputed using Forecaster's
            `set_out_sample_residuals()` method. Defaults to True.
        use_binned_residuals (bool, optional): If `True`, residuals are selected
            based on the predicted values (binned selection).
            If `False`, residuals are selected randomly. Defaults to True.
        random_state (int, optional): Seed for the random number generator to
            ensure reproducibility. Defaults to 123.
        return_predictors (bool, optional): If `True`, the predictors used to make
            the predictions are also returned. Defaults to False.
        n_jobs (int | str, optional): The number of jobs to run in parallel. If `-1`,
            then the number of jobs is set to the number of cores. If 'auto', `n_jobs`
            is set using the function `skforecast.utils.select_n_jobs_backtesting`.
            Defaults to 'auto'.
        verbose (bool, optional): Print number of folds and index of training and
            validation sets used for backtesting. Defaults to False.
        show_progress (bool, optional): Whether to show a progress bar.
            Defaults to True.
        suppress_warnings (bool, optional): If `True`, spotforecast warnings will be
            suppressed during the backtesting process. See
            `spotforecast.exceptions.warn_skforecast_categories` for more information.
            Defaults to False.

    Returns:
        tuple (pd.DataFrame, pd.DataFrame):
            - metric_values: Value(s) of the metric(s).
            - backtest_predictions: Value of predictions. The DataFrame includes
              the following columns:

              - fold: Indicates the fold number where the prediction was made.
              - pred: Predicted values for the corresponding series and time steps.

              If `interval` is not `None`, additional columns are included depending
              on the method:

              - For `float`: Columns `lower_bound` and `upper_bound`.
              - For `list` or `tuple` of 2 elements: Columns `lower_bound` and
                `upper_bound`.
              - For `list` or `tuple` with multiple percentiles: One column per
                percentile (e.g., `p_10`, `p_50`, `p_90`).
              - For `'bootstrapping'`: One column per bootstrapping iteration
                (e.g., `pred_boot_0`, `pred_boot_1`, ..., `pred_boot_n`).
              - For `scipy.stats` distribution objects: One column for each
                estimated parameter of the distribution (e.g., `loc`, `scale`).

              If `return_predictors` is `True`, one column per predictor is created.

              Depending on the relation between `steps` and `fold_stride`, the output
              may include repeated indexes (if `fold_stride < steps`) or gaps
              (if `fold_stride > steps`). See Notes below for more details.

    Notes:
        Note on `fold_stride` vs. `steps`:

        - If `fold_stride == steps`, test sets are placed back-to-back without overlap.
          Each observation appears only once in the output DataFrame, so the
          index is unique.
        - If `fold_stride < steps`, test sets overlap. Multiple forecasts are
          generated for the same observations and, therefore, the output
          DataFrame contains repeated indexes.
        - If `fold_stride > steps`, there are gaps between consecutive test sets.
          Some observations in the series will not have associated predictions,
          so the output DataFrame has non-contiguous indexes.

    References:
        .. [1] Forecasting: Principles and Practice (3rd ed) Rob J Hyndman and
               George Athanasopoulos. https://otexts.com/fpp3/prediction-intervals.html
        .. [2] MAPIE - Model Agnostic Prediction Interval Estimator.
               https://mapie.readthedocs.io/en/stable/theoretical_description_regression.html#the-split-method
    """

    set_skforecast_warnings(suppress_warnings, action="ignore")

    forecaster = deepcopy(forecaster)
    is_regression = forecaster.__spotforecast_tags__["forecaster_task"] == "regression"
    cv = deepcopy(cv)

    cv.set_params(
        {
            "window_size": forecaster.window_size,
            "differentiation": forecaster.differentiation_max,
            "return_all_indexes": False,
            "verbose": verbose,
        }
    )

    refit = cv.refit
    overlapping_folds = cv.overlapping_folds

    if n_jobs == "auto":
        n_jobs = select_n_jobs_backtesting(forecaster=forecaster, refit=refit)
    elif not isinstance(refit, bool) and refit != 1 and n_jobs != 1:
        warnings.warn(
            "If `refit` is an integer other than 1 (intermittent refit). `n_jobs` "
            "is set to 1 to avoid unexpected results during parallelization.",
            IgnoredArgumentWarning,
        )
        n_jobs = 1
    else:
        n_jobs = n_jobs if n_jobs > 0 else cpu_count()

    if not isinstance(metric, list):
        metrics = [
            (
                _get_metric(metric=metric)
                if isinstance(metric, str)
                else add_y_train_argument(metric)
            )
        ]
    else:
        metrics = [
            _get_metric(metric=m) if isinstance(m, str) else add_y_train_argument(m)
            for m in metric
        ]

    store_in_sample_residuals = True if use_in_sample_residuals else False
    if interval is None:
        forecaster._probabilistic_mode = False
    elif use_binned_residuals:
        forecaster._probabilistic_mode = "binned"
    else:
        forecaster._probabilistic_mode = "no_binned"

    folds = cv.split(X=y, as_pandas=False)
    initial_train_size = cv.initial_train_size
    window_size = cv.window_size
    gap = cv.gap

    if initial_train_size is not None:
        # NOTE: This allows for parallelization when `refit` is `False`. The initial
        # Forecaster fit occurs outside of the auxiliary function.
        exog_train = exog.iloc[:initial_train_size,] if exog is not None else None
        forecaster.fit(
            y=y.iloc[:initial_train_size,],
            exog=exog_train,
            store_in_sample_residuals=store_in_sample_residuals,
        )
        folds[0][5] = False

    if refit:
        n_of_fits = int(len(folds) / refit)
        if type(forecaster).__name__ != "ForecasterDirect" and n_of_fits > 50:
            warnings.warn(
                f"The forecaster will be fit {n_of_fits} times. This can take substantial"
                f" amounts of time. If not feasible, try with `refit = False`.\n",
                LongTrainingWarning,
            )
        elif (
            type(forecaster).__name__ == "ForecasterDirect"
            and n_of_fits * forecaster.max_step > 50
        ):
            warnings.warn(
                f"The forecaster will be fit {n_of_fits * forecaster.max_step} times "
                f"({n_of_fits} folds * {forecaster.max_step} estimators). This can take "
                f"substantial amounts of time. If not feasible, try with `refit = False`.\n",
                LongTrainingWarning,
            )

    if show_progress:
        folds = tqdm(folds)

    def _fit_predict_forecaster(
        fold,
        forecaster,
        y,
        exog,
        store_in_sample_residuals,
        gap,
        interval,
        interval_method,
        n_boot,
        use_in_sample_residuals,
        use_binned_residuals,
        random_state,
        return_predictors,
        is_regression,
    ) -> pd.DataFrame:
        """
        Fit the forecaster and predict `steps` ahead. This is an auxiliary
        function used to parallelize the backtesting_forecaster function.
        """

        train_iloc_start = fold[1][0]
        train_iloc_end = fold[1][1]
        last_window_iloc_start = fold[2][0]
        last_window_iloc_end = fold[2][1]
        test_iloc_start = fold[3][0]
        test_iloc_end = fold[3][1]

        if fold[5] is False:
            # When the model is not fitted, last_window must be updated to include
            # the data needed to make predictions.
            last_window_y = y.iloc[last_window_iloc_start:last_window_iloc_end]
        else:
            # The model is fitted before making predictions. If `fixed_train_size`
            # the train size doesn't increase but moves by `steps` in each iteration.
            # If `False` the train size increases by `steps` in each iteration.
            y_train = y.iloc[train_iloc_start:train_iloc_end,]
            exog_train = (
                exog.iloc[train_iloc_start:train_iloc_end,]
                if exog is not None
                else None
            )
            last_window_y = None
            forecaster.fit(
                y=y_train,
                exog=exog_train,
                store_in_sample_residuals=store_in_sample_residuals,
            )

        next_window_exog = (
            exog.iloc[test_iloc_start:test_iloc_end,] if exog is not None else None
        )

        steps = len(range(test_iloc_start, test_iloc_end))
        if type(forecaster).__name__ == "ForecasterDirect" and gap > 0:
            # Select only the steps that need to be predicted if gap > 0
            test_no_gap_iloc_start = fold[4][0]
            test_no_gap_iloc_end = fold[4][1]
            n_steps = test_no_gap_iloc_end - test_no_gap_iloc_start
            steps = list(range(gap + 1, gap + 1 + n_steps))

        preds = []
        if is_regression:
            if interval is not None:
                kwargs_interval = {
                    "steps": steps,
                    "last_window": last_window_y,
                    "exog": next_window_exog,
                    "n_boot": n_boot,
                    "use_in_sample_residuals": use_in_sample_residuals,
                    "use_binned_residuals": use_binned_residuals,
                    "random_state": random_state,
                }
                if interval_method == "bootstrapping":
                    if interval == "bootstrapping":
                        # Return all bootstrap samples
                        pred = forecaster.predict_bootstrapping(**kwargs_interval)
                    else:
                        # Use predict_interval with bootstrapping method
                        pred = forecaster.predict_interval(
                            method="bootstrapping", interval=interval, **kwargs_interval
                        )
                    preds.append(pred)
                else:
                    # Use conformal method
                    pred = forecaster.predict_interval(
                        method="conformal", interval=interval, **kwargs_interval
                    )
                    preds.append(pred)

            # NOTE: This is done after probabilistic predictions to avoid repeating
            # the same checks.
            if interval is None or interval_method != "conformal":
                pred = forecaster.predict(
                    steps=steps,
                    last_window=last_window_y,
                    exog=next_window_exog,
                    check_inputs=True if interval is None else False,
                )
                preds.insert(0, pred)
        else:
            pred = forecaster.predict_proba(
                steps=steps, last_window=last_window_y, exog=next_window_exog
            )
            preds.append(pred)

        if return_predictors:
            pred = forecaster.create_predict_X(
                steps=steps,
                last_window=last_window_y,
                exog=next_window_exog,
                check_inputs=False,
            )
            preds.append(pred)

        if len(preds) == 1:
            pred = preds[0]
        else:
            pred = pd.concat(preds, axis=1)

        if type(forecaster).__name__ != "ForecasterDirect" and gap > 0:
            pred = pred.iloc[gap:,]

        return pred

    kwargs_fit_predict_forecaster = {
        "forecaster": forecaster,
        "y": y,
        "exog": exog,
        "store_in_sample_residuals": store_in_sample_residuals,
        "gap": gap,
        "interval": interval,
        "interval_method": interval_method,
        "n_boot": n_boot,
        "use_in_sample_residuals": use_in_sample_residuals,
        "use_binned_residuals": use_binned_residuals,
        "random_state": random_state,
        "return_predictors": return_predictors,
        "is_regression": is_regression,
    }
    backtest_predictions = Parallel(n_jobs=n_jobs)(
        delayed(_fit_predict_forecaster)(fold=fold, **kwargs_fit_predict_forecaster)
        for fold in folds
    )
    fold_labels = [
        np.repeat(fold[0], backtest_predictions[i].shape[0])
        for i, fold in enumerate(folds)
    ]

    backtest_predictions = pd.concat(backtest_predictions)
    if isinstance(backtest_predictions, pd.Series):
        backtest_predictions = backtest_predictions.to_frame()

    if not is_regression:
        proba_cols = [f"{cls}_proba" for cls in forecaster.classes_]
        idx_max = backtest_predictions[proba_cols].to_numpy().argmax(axis=1)
        backtest_predictions.insert(0, "pred", np.array(forecaster.classes_)[idx_max])

    backtest_predictions.insert(0, "fold", np.concatenate(fold_labels))

    train_indexes = []
    for i, fold in enumerate(folds):
        fit_fold = fold[-1]
        if i == 0 or fit_fold:
            # NOTE: When using a scaled metric, `y_train` doesn't include the
            # first window_size observations used to create the predictors and/or
            # rolling features.
            train_iloc_start = fold[1][0] + window_size
            train_iloc_end = fold[1][1]
            train_indexes.append(np.arange(train_iloc_start, train_iloc_end))

    train_indexes = np.unique(np.concatenate(train_indexes))
    y_train = y.iloc[train_indexes]

    backtest_predictions_for_metrics = backtest_predictions
    if overlapping_folds:
        backtest_predictions_for_metrics = backtest_predictions_for_metrics.loc[
            ~backtest_predictions_for_metrics.index.duplicated(keep="last")
        ]

    y_true = y.loc[backtest_predictions_for_metrics.index]
    y_pred = backtest_predictions_for_metrics["pred"]
    metric_values = [
        [m(y_true=y_true, y_pred=y_pred, y_train=y_train) for m in metrics]
    ]

    metric_values = pd.DataFrame(
        data=metric_values, columns=[m.__name__ for m in metrics]
    )

    set_skforecast_warnings(suppress_warnings, action="default")

    return metric_values, backtest_predictions


def backtesting_forecaster_one_step(
    forecaster: object,
    y: pd.Series,
    cv: OneStepAheadFold,
    metric: str | Callable | list[str | Callable],
    exog: pd.Series | pd.DataFrame | None = None,
    interval: float | list[float] | tuple[float] | str | object | None = None,
    interval_method: str = "bootstrapping",
    n_boot: int = 250,
    use_in_sample_residuals: bool = True,
    use_binned_residuals: bool = True,
    random_state: int = 123,
    return_predictors: bool = False,
    n_jobs: int | str = "auto",
    verbose: bool = False,
    show_progress: bool = True,
    suppress_warnings: bool = False,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Backtesting of forecaster model using one-step-ahead predictions.

    Args:
        forecaster (ForecasterRecursive, ForecasterDirect, ForecasterEquivalentDate):
            Forecaster model.
        y (pd.Series): Training time series.
        cv (OneStepAheadFold): OneStepAheadFold object with the information needed to
            split the data into folds.
        metric (str | Callable | list): Metric used to quantify the goodness of fit
            of the model.
        exog (pd.Series | pd.DataFrame, optional): Exogenous variable/s included as
            predictor/s. Defaults to None.
        interval (float | list | tuple | str | object, optional): Specifies whether
            probabilistic predictions should be estimated.
        interval_method (str, optional): Technique used to estimate prediction intervals.
        n_boot (int, optional): Number of bootstrapping iterations.
        use_in_sample_residuals (bool, optional): Use residuals from training data.
        use_binned_residuals (bool, optional): Use binned residuals for intervals.
        random_state (int, optional): Seed for reproducibility.
        return_predictors (bool, optional): Return predictors used for each prediction.
        n_jobs (int | str, optional): Number of jobs to run in parallel.
        verbose (bool, optional): Print information about the process.
        show_progress (bool, optional): Whether to show a progress bar.
        suppress_warnings (bool, optional): Suppress spotforecast warnings.

    Returns:
        tuple (pd.DataFrame, pd.DataFrame):
            - metric_values: Value(s) of the metric(s).
            - backtest_predictions: Value of predictions.

    Notes:
        This function is designed for one-step-ahead backtesting,
        where predictions are made for the next time step using the most recent data. The function handles the fitting and
        prediction process for each fold defined in the `OneStepAheadFold` object,
        and calculates the specified metric(s) based on the true and predicted values.
        Depending on the `interval` and `interval_method` parameters, it can also generate probabilistic predictions.

    Examples:
        >>> from spotforecast2_safe.forecaster import ForecasterRecursive
        >>> from spotforecast2_safe.model_selection.split_one_step import OneStepAheadFold
        >>> from spotforecast2_safe.model_selection.validation import backtesting_forecaster_one_step
        >>> # Create a forecaster and a one-step-ahead fold
        >>> import numpy as np
        >>> import pandas as pd
        >>> from sklearn.linear_model import LinearRegression
        >>> y = pd.Series(np.random.randn(100), name='y')
        >>> forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=5)
        >>> cv = OneStepAheadFold(initial_train_size=20, window_size=5)
        >>> # Perform backtesting
        >>> metric_values, backtest_predictions = backtesting_forecaster_one_step(
        ...     forecaster=forecaster,
        ...     y=y,
        ...     cv=cv,
        ...     metric='mean_squared_error',
        ...     exog=None,
        ...     interval=0.95,
        ...     interval_method='bootstrapping',
        ...     n_boot=20,
        ...     use_in_sample_residuals=True,
        ...     use_binned_residuals=False,
        ...     random_state=42,
        ...     return_predictors=False,
        ...     n_jobs=1,
        ...     verbose=True,
        ...     show_progress=True,
        ...     suppress_warnings=False
        ... )
        # Note: For reliable bootstrapping with binned residuals, use a sufficiently large series and value spread.
        # For random data, use_binned_residuals=False.
        # TODO: Setting return_predictors=True requires ForecasterRecursive.create_predict_X().

    """

    set_skforecast_warnings(suppress_warnings, action="ignore")

    forecaster = deepcopy(forecaster)
    cv = deepcopy(cv)

    check_one_step_ahead_input(
        forecaster=forecaster,
        cv=cv,
        y=y,
        metric=metric,
        exog=exog,
        show_progress=show_progress,
        suppress_warnings=suppress_warnings,
    )

    cv.set_params(
        {
            "window_size": forecaster.window_size,
            "differentiation": forecaster.differentiation_max,
            "verbose": verbose,
        }
    )

    if not isinstance(metric, list):
        metrics = [
            (
                _get_metric(metric=metric)
                if isinstance(metric, str)
                else add_y_train_argument(metric)
            )
        ]
    else:
        metrics = [
            _get_metric(metric=m) if isinstance(m, str) else add_y_train_argument(m)
            for m in metric
        ]

    # Split data to get training set for fitting
    X_train, y_train, X_test, y_test = forecaster._train_test_split_one_step_ahead(
        y=y, initial_train_size=cv.initial_train_size, exog=exog
    )

    # Fit the forecaster
    forecaster.fit(
        y=y_train,
        exog=exog.iloc[: cv.initial_train_size] if exog is not None else None,
        store_in_sample_residuals=use_in_sample_residuals,
    )

    # Predictions
    # NOTE: For one-step-ahead, we can use the estimator directly on X_test
    # which contains all necessary lags and exog features.
    if interval is not None:
        if interval_method == "bootstrapping":
            if interval == "bootstrapping":
                backtest_predictions = forecaster.predict_bootstrapping(
                    steps=len(X_test),
                    last_window=None,
                    exog=(
                        exog.iloc[cv.initial_train_size :] if exog is not None else None
                    ),
                    n_boot=n_boot,
                    use_in_sample_residuals=use_in_sample_residuals,
                    use_binned_residuals=use_binned_residuals,
                    random_state=random_state,
                    in_sample_predict=True,  # This might be needed if supported
                )
            else:
                backtest_predictions = forecaster.predict_interval(
                    steps=len(X_test),
                    last_window=None,
                    exog=(
                        exog.iloc[cv.initial_train_size :] if exog is not None else None
                    ),
                    method="bootstrapping",
                    interval=interval,
                    n_boot=n_boot,
                    use_in_sample_residuals=use_in_sample_residuals,
                    use_binned_residuals=use_binned_residuals,
                    random_state=random_state,
                )
        else:
            backtest_predictions = forecaster.predict_interval(
                steps=len(X_test),
                last_window=None,
                exog=exog.iloc[cv.initial_train_size :] if exog is not None else None,
                method="conformal",
                interval=interval,
                n_boot=n_boot,
                use_in_sample_residuals=use_in_sample_residuals,
                use_binned_residuals=use_binned_residuals,
                random_state=random_state,
            )

        # Add 'pred' column
        if "pred" not in backtest_predictions.columns:
            pred = forecaster.estimator.predict(X_test).ravel()
            backtest_predictions.insert(0, "pred", pred)
    else:
        pred = forecaster.estimator.predict(X_test).ravel()
        backtest_predictions = pd.DataFrame({"pred": pred}, index=y_test.index)

    if return_predictors:
        X_predict = forecaster.create_predict_X(
            steps=len(X_test), last_window=None, exog=X_test
        )
        backtest_predictions = pd.concat([backtest_predictions, X_predict], axis=1)

    backtest_predictions.insert(0, "fold", 0)

    # Calculate metrics
    y_true = y_test
    # Metrics expect y_true, y_pred, y_train
    # y_train should not include first window_size observations
    y_train_metrics = y_train.iloc[forecaster.window_size :]

    metric_values = []
    for m in metrics:
        metric_values.append(
            m(
                y_true=y_true,
                y_pred=backtest_predictions["pred"],
                y_train=y_train_metrics,
            )
        )

    metric_values = pd.DataFrame(
        data=[metric_values], columns=[m.__name__ for m in metrics]
    )

    set_skforecast_warnings(suppress_warnings, action="default")

    return metric_values, backtest_predictions


def backtesting_forecaster(
    forecaster: object,
    y: pd.Series,
    cv: TimeSeriesFold,
    metric: str | Callable | list[str | Callable],
    exog: pd.Series | pd.DataFrame | None = None,
    interval: float | list[float] | tuple[float] | str | object | None = None,
    interval_method: str = "bootstrapping",
    n_boot: int = 250,
    use_in_sample_residuals: bool = True,
    use_binned_residuals: bool = True,
    random_state: int = 123,
    return_predictors: bool = False,
    n_jobs: int | str = "auto",
    verbose: bool = False,
    show_progress: bool = True,
    suppress_warnings: bool = False,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Backtesting of forecaster model following the folds generated by the TimeSeriesFold
    class and using the metric(s) provided.

    If `forecaster` is already trained and `initial_train_size` is set to `None` in the
    TimeSeriesFold class, no initial train will be done and all data will be used
    to evaluate the model. However, the first `len(forecaster.last_window)` observations
    are needed to create the initial predictors, so no predictions are calculated for
    them.

    A copy of the original forecaster is created so that it is not modified during
    the process.

    Args:
        forecaster (ForecasterRecursive, ForecasterDirect, ForecasterEquivalentDate):
            Forecaster model.
        y (pd.Series): Training time series.
        cv (TimeSeriesFold): TimeSeriesFold object with the information needed to
            split the data into folds.
        metric (str | Callable | list): Metric used to quantify the goodness of fit
            of the model.

            - If `str`: {'mean_squared_error', 'mean_absolute_error',
              'mean_absolute_percentage_error', 'mean_squared_log_error',
              'mean_absolute_scaled_error', 'root_mean_squared_scaled_error'}
            - If `Callable`: Function with arguments `y_true`, `y_pred` and `y_train`
              (Optional) that returns a float.
            - If `list`: List containing multiple strings and/or Callables.
        exog (pd.Series | pd.DataFrame, optional): Exogenous variable/s included as
            predictor/s. Must have the same number of observations as `y` and should
            be aligned so that y[i] is regressed on exog[i]. Defaults to None.
        interval (float | list | tuple | str | object, optional): Specifies whether
            probabilistic predictions should be estimated and the method to use.
            The following options are supported:

            - If `float`, represents the nominal (expected) coverage (between 0 and 1).
            For instance, `interval=0.95` corresponds to `[2.5, 97.5]` percentiles.
            - If `list` or `tuple`: Sequence of percentiles to compute, each value must
              be between 0 and 100 inclusive. For example, a 95% confidence interval can
              be specified as `interval = [2.5, 97.5]` or multiple percentiles (e.g. 10,
              50 and 90) as `interval = [10, 50, 90]`.
            - If 'bootstrapping' (str): `n_boot` bootstrapping predictions will be
              generated.
            - If scipy.stats distribution object, the distribution parameters will
              be estimated for each prediction.
            - If None, no probabilistic predictions are estimated.
            Defaults to None.
        interval_method (str, optional): Technique used to estimate prediction
            intervals. Available options:

            - 'bootstrapping': Bootstrapping is used to generate prediction intervals.
            - 'conformal': Employs the conformal prediction split method for
              interval estimation.
            Defaults to 'bootstrapping'.
        n_boot (int, optional): Number of bootstrapping iterations to perform when
            estimating prediction intervals. Defaults to 250.
        use_in_sample_residuals (bool, optional): If `True`, residuals from the
            training data are used as proxy of prediction error to create predictions.
            If `False`, out of sample residuals (calibration) are used.
            Out-of-sample residuals must be precomputed using Forecaster's
            `set_out_sample_residuals()` method. Defaults to True.
        use_binned_residuals (bool, optional): If `True`, residuals are selected
            based on the predicted values (binned selection).
            If `False`, residuals are selected randomly. Defaults to True.
        random_state (int, optional): Seed for the random number generator to
            ensure reproducibility. Defaults to 123.
        return_predictors (bool, optional): If `True`, the predictors used to make
            the predictions are also returned. Defaults to False.
        n_jobs (int | str, optional): The number of jobs to run in parallel. If `-1`,
            then the number of jobs is set to the number of cores. If 'auto', `n_jobs`
            is set using the function `skforecast.utils.select_n_jobs_backtesting`.
            Defaults to 'auto'.
        verbose (bool, optional): Print number of folds and index of training and
            validation sets used for backtesting. Defaults to False.
        show_progress (bool, optional): Whether to show a progress bar.
            Defaults to True.
        suppress_warnings (bool, optional): If `True`, spotforecast warnings will be
            suppressed during the backtesting process. See
            `spotforecast.exceptions.warn_skforecast_categories` for more information.
            Defaults to False.

    Returns:
        tuple (pd.DataFrame, pd.DataFrame):
            - metric_values: Value(s) of the metric(s).
            - backtest_predictions: Value of predictions. The DataFrame includes
              the following columns:

              - fold: Indicates the fold number where the prediction was made.
              - pred: Predicted values for the corresponding series and time steps.

              If `interval` is not `None`, additional columns are included depending
              on the method:

              - For `float`: Columns `lower_bound` and `upper_bound`.
              - For `list` or `tuple` of 2 elements: Columns `lower_bound` and
                `upper_bound`.
              - For `list` or `tuple` with multiple percentiles: One column per
                percentile (e.g., `p_10`, `p_50`, `p_90`).
              - For `'bootstrapping'`: One column per bootstrapping iteration
                (e.g., `pred_boot_0`, `pred_boot_1`, ..., `pred_boot_n`).
              - For `scipy.stats` distribution objects: One column for each
                estimated parameter of the distribution (e.g., `loc`, `scale`).

              If `return_predictors` is `True`, one column per predictor is created.

              Depending on the relation between `steps` and `fold_stride`, the output
              may include repeated indexes (if `fold_stride < steps`) or gaps
              (if `fold_stride > steps`). See Notes below for more details.

    Notes:
        Note on `fold_stride` vs. `steps`:

        - If `fold_stride == steps`, test sets are placed back-to-back without overlap.
          Each observation appears only once in the output DataFrame, so the
          index is unique.
        - If `fold_stride < steps`, test sets overlap. Multiple forecasts are
          generated for the same observations and, therefore, the output
          DataFrame contains repeated indexes.
        - If `fold_stride > steps`, there are gaps between consecutive test sets.
          Some observations in the series will not have associated predictions,
          so the output DataFrame has non-contiguous indexes.

    Examples:
        >>> import pandas as pd
        >>> from sklearn.ensemble import RandomForestRegressor
        >>> from spotforecast2_safe.forecaster.recursive import ForecasterRecursive
        >>> from spotforecast2_safe.model_selection import backtesting_forecaster, TimeSeriesFold
        >>> y = pd.Series([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        >>> forecaster = ForecasterRecursive(
        ...     estimator=RandomForestRegressor(random_state=123),
        ...     lags=5
        ... )
        >>> cv = TimeSeriesFold(
        ...     steps=2,
        ...     initial_train_size=5,
        ...     refit=False
        ... )
        >>> metric_values, backtest_predictions = backtesting_forecaster(
        ...     forecaster=forecaster,
        ...     y=y,
        ...     cv=cv,
        ...     metric='mean_squared_error'
        ... )
        >>> metric_values
           mean_squared_error
        0            0.201334
        >>> backtest_predictions
           fold  pred
        5     0  5.18
        6     0  6.10
        7     1  7.36
        8     1  8.40
        9     2  9.31
    """

    if type(cv).__name__ == "OneStepAheadFold":
        return backtesting_forecaster_one_step(
            forecaster=forecaster,
            y=y,
            cv=cv,
            metric=metric,
            exog=exog,
            interval=interval,
            interval_method=interval_method,
            n_boot=n_boot,
            use_in_sample_residuals=use_in_sample_residuals,
            use_binned_residuals=use_binned_residuals,
            random_state=random_state,
            return_predictors=return_predictors,
            n_jobs=n_jobs,
            verbose=verbose,
            show_progress=show_progress,
            suppress_warnings=suppress_warnings,
        )

    check_backtesting_input(
        forecaster=forecaster,
        cv=cv,
        y=y,
        metric=metric,
        interval=interval,
        interval_method=interval_method,
        n_boot=n_boot,
        use_in_sample_residuals=use_in_sample_residuals,
        use_binned_residuals=use_binned_residuals,
        random_state=random_state,
        return_predictors=return_predictors,
        n_jobs=n_jobs,
        show_progress=show_progress,
        suppress_warnings=suppress_warnings,
    )

    metric_values, backtest_predictions = _backtesting_forecaster(
        forecaster=forecaster,
        y=y,
        cv=cv,
        metric=metric,
        exog=exog,
        interval=interval,
        interval_method=interval_method,
        n_boot=n_boot,
        use_in_sample_residuals=use_in_sample_residuals,
        use_binned_residuals=use_binned_residuals,
        random_state=random_state,
        return_predictors=return_predictors,
        n_jobs=n_jobs,
        verbose=verbose,
        show_progress=show_progress,
        suppress_warnings=suppress_warnings,
    )

    return metric_values, backtest_predictions



# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later

from .task_safe_demo import main as demo_main
from .task_safe_n_to_1_with_covariates_and_dataframe import (
    main as n2o1_cov_df_main,
)

__all__ = ["demo_main", "n2o1_cov_df_main"]



# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later

"""Repeating Basis Function transformer for cyclical features."""

from typing import Any, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class RepeatingBasisFunction(BaseEstimator, TransformerMixin):
    """
    Transformer that encodes cyclical features using repeating radial basis functions.

    This transformer places Gaussian basis functions across the specified input range
    and wraps them around to handle periodicity (e.g., day of year, hour of day).
    It is a simplified implementation to avoid external dependencies like scikit-lego.

    Attributes:
        n_periods (int): Number of basis functions to place.
        column (str): Name of the column in the input DataFrame/Series to transform.
        input_range (Tuple[int, int]): The range of the input values (min, max).
        remainder (str): Policy for remaining columns (currently only 'drop' is supported).

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> from spotforecast2_safe.preprocessing.repeating_basis_function import RepeatingBasisFunction
        >>> X = pd.DataFrame({"hour": [0, 6, 12, 18, 23]})
        >>> rbf = RepeatingBasisFunction(n_periods=4, column="hour", input_range=(0, 23))
        >>> features = rbf.fit_transform(X)
        >>> features.shape
        (5, 4)
    """

    def __init__(
        self,
        n_periods: int,
        column: str,
        input_range: Tuple[int, int],
        remainder: str = "drop",
    ):
        """
        Initialize the RepeatingBasisFunction transformer.

        Args:
            n_periods: Number of basis functions.
            column: Name of the column to transform.
            input_range: Min and max values of the periodic feature.
            remainder: How to handle other columns. Defaults to "drop".
        """
        self.n_periods = n_periods
        self.column = column
        self.input_range = input_range
        self.remainder = remainder

    def fit(self, X: Any, y: Any = None) -> "RepeatingBasisFunction":
        """
        Fitted transformer (no-op).

        Args:
            X: Input data.
            y: Ignored.

        Returns:
            self: The fitted transformer.
        """
        return self

    def transform(self, X: Union[pd.Series, pd.DataFrame]) -> np.ndarray:
        """
        Transform the input data into RBF features.

        Args:
            X: Input DataFrame or Series containing the column to transform.

        Returns:
            np.ndarray: Array of transformed features with shape (n_samples, n_periods).

        Raises:
            ValueError: If the specified column is not found in the input.
        """
        # Allow passing just the column series if X is not a DataFrame
        if isinstance(X, pd.Series):
            vals = X.values
        elif isinstance(X, pd.DataFrame) and self.column in X.columns:
            vals = X[self.column].values
        else:
            raise ValueError(f"Column {self.column} not found in input")

        # Normalize to [0, 1] relative to input range
        vals_norm = (vals - self.input_range[0]) / (
            self.input_range[1] - self.input_range[0]
        )

        features = []
        for i in range(self.n_periods):
            mu = i / self.n_periods
            # Gaussian with wraparound handling for cyclic
            diff = np.abs(vals_norm - mu)
            diff = np.minimum(diff, 1 - diff)  # cyclic distance
            # sigma estimated as 1 / n_periods for reasonable overlap
            sigma = 1 / self.n_periods
            val = np.exp(-(diff**2) / (2 * sigma**2))
            features.append(val)

        return np.stack(features, axis=1)



# SPDX-FileCopyrightText: skforecast team
# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later AND BSD-3-Clause

"""
Time series cross-validation splitting.
"""

from __future__ import annotations
import warnings
import itertools
import numpy as np
import pandas as pd

from spotforecast2_safe.forecaster.utils import (
    date_to_index_position,
    get_style_repr_html,
)
from spotforecast2_safe.exceptions import IgnoredArgumentWarning
from .split_base import BaseFold


class TimeSeriesFold(BaseFold):
    """Class to split time series data into train and test folds.

    When used within a backtesting or hyperparameter search, the arguments
    'initial_train_size', 'window_size' and 'differentiation' are not required
    as they are automatically set by the backtesting or hyperparameter search
    functions.

    Args:
        steps: Number of observations used to be predicted in each fold.
            This is also commonly referred to as the forecast horizon or test size.
        initial_train_size: Number of observations used for initial training.

            - If `None` or 0, the initial forecaster is not trained in the first fold.
            - If an integer, the number of observations used for initial training.
            - If a date string or pandas Timestamp, it is the last date included in
              the initial training set.

            Defaults to None.
        fold_stride: Number of observations that the start of the test set
            advances between consecutive folds.

            - If `None`, it defaults to the same value as `steps`, meaning that folds
              are placed back-to-back without overlap.
            - If `fold_stride < steps`, test sets overlap and multiple forecasts will
              be generated for the same observations.
            - If `fold_stride > steps`, gaps are left between consecutive test sets.

            Defaults to None.
        window_size: Number of observations needed to generate the
            autoregressive predictors. Defaults to None.
        differentiation: Number of observations to use for differentiation.
            This is used to extend the `last_window` as many observations as the
            differentiation order. Defaults to None.
        refit: Whether to refit the forecaster in each fold.

            - If `True`, the forecaster is refitted in each fold.
            - If `False`, the forecaster is trained only in the first fold.
            - If an integer, the forecaster is trained in the first fold and then refitted
              every `refit` folds.

            Defaults to False.
        fixed_train_size: Whether the training size is fixed or increases
            in each fold. Defaults to True.
        gap: Number of observations between the end of the training set
            and the start of the test set. Defaults to 0.
        skip_folds: Number of folds to skip.

            - If an integer, every 'skip_folds'-th is returned.
            - If a list, the indexes of the folds to skip.

            For example, if `skip_folds=3` and there are 10 folds, the returned folds are
            0, 3, 6, and 9. If `skip_folds=[1, 2, 3]`, the returned folds are 0, 4, 5, 6, 7,
            8, and 9. Defaults to None.
        allow_incomplete_fold: Whether to allow the last fold to include
            fewer observations than `steps`. If `False`, the last fold is excluded if it
            is incomplete. Defaults to True.
        return_all_indexes: Whether to return all indexes or only the
            start and end indexes of each fold. Defaults to False.
        verbose: Whether to print information about generated folds.
            Defaults to True.

    Attributes:
        steps: Number of observations used to be predicted in each fold.
        initial_train_size: Number of observations used for initial training.
            If `None` or 0, the initial forecaster is not trained in the first fold.
        fold_stride: Number of observations that the start of the test set
            advances between consecutive folds.
        overlapping_folds: Whether the folds overlap.
        window_size: Number of observations needed to generate the
            autoregressive predictors.
        differentiation: Number of observations to use for differentiation.
            This is used to extend the `last_window` as many observations as the
            differentiation order.
        refit: Whether to refit the forecaster in each fold.
        fixed_train_size: Whether the training size is fixed or increases in each fold.
        gap: Number of observations between the end of the training set and the
            start of the test set.
        skip_folds: Number of folds to skip.
        allow_incomplete_fold: Whether to allow the last fold to include fewer
            observations than `steps`.
        return_all_indexes: Whether to return all indexes or only the start
            and end indexes of each fold.
        verbose: Whether to print information about generated folds.

    Examples:
        Basic usage with fixed train size:
        >>> import pandas as pd
        >>> import numpy as np
        >>> from spotforecast2_safe.model_selection import TimeSeriesFold
        >>> # Create sample time series data
        >>> dates = pd.date_range('2020-01-01', periods=100, freq='D')
        >>> y = pd.Series(np.arange(100), index=dates)
        >>> # Create fold splitter
        >>> cv = TimeSeriesFold(
        ...     steps=10,
        ...     initial_train_size=50,
        ...     refit=True,
        ...     fixed_train_size=True
        ... )
        >>> # Get folds
        >>> folds = cv.split(y)
        >>> print(f"Number of folds: {len(folds)}")
        Number of folds: 4

        Overlapping folds with custom stride:
        >>> cv = TimeSeriesFold(
        ...     steps=30,
        ...     initial_train_size=50,
        ...     fold_stride=7,
        ...     fixed_train_size=False
        ... )
        >>> folds = cv.split(y)
        >>> # First test fold covers [50, 80), second [57, 87), etc.

        Return as pandas DataFrame:
        >>> cv = TimeSeriesFold(steps=10, initial_train_size=50)
        >>> folds_df = cv.split(y, as_pandas=True)
        >>> print(folds_df.columns.tolist())
        ['fold', 'train_start', 'train_end', 'last_window_start', 'last_window_end', 'test_start', 'test_end', 'test_start_with_gap', 'test_end_with_gap', 'fit_forecaster']

        Skip folds for faster evaluation:
        >>> cv = TimeSeriesFold(
        ...     steps=5,
        ...     initial_train_size=50,
        ...     skip_folds=2
        ... )
        >>> folds = cv.split(y)
        >>> # Returns folds 0, 2, 4, 6, ...

    Note:
        Returned values are the positions of the observations and not the actual values of
        the index, so they can be used to slice the data directly using iloc. For example,
        if the input series is `X = [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]`, the
        `initial_train_size = 3`, `window_size = 2`, `steps = 4`, and `gap = 1`,
        the output of the first fold will: [0, [0, 3], [1, 3], [3, 8], [4, 8], True].

        The first element is the fold number, the first list `[0, 3]` indicates that
        the training set goes from the first to the third observation. The second
        list `[1, 3]` indicates that the last window seen by the forecaster during
        training goes from the second to the third observation. The third list `[3, 8]`
        indicates that the test set goes from the fourth to the eighth observation.
        The fourth list `[4, 8]` indicates that the test set including the gap goes
        from the fifth to the eighth observation. The boolean `False` indicates that
        the forecaster should not be trained in this fold.

        Following the python convention, the start index is inclusive and the end index is
        exclusive. This means that the last index is not included in the slice.

        As an example, with `initial_train_size=50`, `steps=30`, and `fold_stride=7`,
        the first test fold will cover observations [50, 80), the second fold [57, 87),
        and the third fold [64, 94). This configuration produces multiple forecasts
        for the same observations, which is often desirable in rolling-origin
        evaluation.
    """

    def __init__(
        self,
        steps: int,
        initial_train_size: int | str | pd.Timestamp | None = None,
        fold_stride: int | None = None,
        window_size: int | None = None,
        differentiation: int | None = None,
        refit: bool | int = False,
        fixed_train_size: bool = True,
        gap: int = 0,
        skip_folds: int | list[int] | None = None,
        allow_incomplete_fold: bool = True,
        return_all_indexes: bool = False,
        verbose: bool = True,
    ) -> None:

        super().__init__(
            steps=steps,
            initial_train_size=initial_train_size,
            fold_stride=fold_stride,
            window_size=window_size,
            differentiation=differentiation,
            refit=refit,
            fixed_train_size=fixed_train_size,
            gap=gap,
            skip_folds=skip_folds,
            allow_incomplete_fold=allow_incomplete_fold,
            return_all_indexes=return_all_indexes,
            verbose=verbose,
        )

        self.steps = steps
        self.fold_stride = fold_stride if fold_stride is not None else steps
        self.overlapping_folds = self.fold_stride < self.steps
        self.refit = refit
        self.fixed_train_size = fixed_train_size
        self.gap = gap
        self.skip_folds = skip_folds
        self.allow_incomplete_fold = allow_incomplete_fold

    def __repr__(self) -> str:
        """Information displayed when printed.

        Args:
            None

        Returns:
            String representation of the TimeSeriesFold object.

        Examples:
            >>> from spotforecast2_safe.model_selection import TimeSeriesFold
            >>> cv = TimeSeriesFold(steps=1)
            >>> print(cv)
        """

        info = (
            f"{'=' * len(type(self).__name__)} \n"
            f"{type(self).__name__} \n"
            f"{'=' * len(type(self).__name__)} \n"
            f"Initial train size    = {self.initial_train_size},\n"
            f"Steps                 = {self.steps},\n"
            f"Fold stride           = {self.fold_stride},\n"
            f"Overlapping folds     = {self.overlapping_folds},\n"
            f"Window size           = {self.window_size},\n"
            f"Differentiation       = {self.differentiation},\n"
            f"Refit                 = {self.refit},\n"
            f"Fixed train size      = {self.fixed_train_size},\n"
            f"Gap                   = {self.gap},\n"
            f"Skip folds            = {self.skip_folds},\n"
            f"Allow incomplete fold = {self.allow_incomplete_fold},\n"
            f"Return all indexes    = {self.return_all_indexes},\n"
            f"Verbose               = {self.verbose}\n"
        )

        return info

    def _repr_html_(self) -> str:
        """HTML representation of the object.

        The "General Information" section is expanded by default.

        Returns:
            HTML string representation for Jupyter notebooks.
        """

        style, unique_id = get_style_repr_html()
        content = f"""
        <div class="container-{unique_id}">
            <p style="font-size: 1.5em; font-weight: bold; margin-block-start: 0.83em; margin-block-end: 0.83em;">{type(self).__name__}</p>
            <details open>
                <summary>General Information</summary>
                <ul>
                    <li><strong>Initial train size:</strong> {self.initial_train_size}</li>
                    <li><strong>Steps:</strong> {self.steps}</li>
                    <li><strong>Fold stride:</strong> {self.fold_stride}</li>
                    <li><strong>Overlapping folds:</strong> {self.overlapping_folds}</li>
                    <li><strong>Window size:</strong> {self.window_size}</li>
                    <li><strong>Differentiation:</strong> {self.differentiation}</li>
                    <li><strong>Refit:</strong> {self.refit}</li>
                    <li><strong>Fixed train size:</strong> {self.fixed_train_size}</li>
                    <li><strong>Gap:</strong> {self.gap}</li>
                    <li><strong>Skip folds:</strong> {self.skip_folds}</li>
                    <li><strong>Allow incomplete fold:</strong> {self.allow_incomplete_fold}</li>
                    <li><strong>Return all indexes:</strong> {self.return_all_indexes}</li>
                </ul>
            </details>
        </div>
        """

        return style + content

    def split(
        self,
        X: pd.Series | pd.DataFrame | pd.Index | dict[str, pd.Series | pd.DataFrame],
        as_pandas: bool = False,
    ) -> list | pd.DataFrame:
        """Split the time series data into train and test folds.

        Args:
            X: Time series data or index to split. Can be a pandas Series, DataFrame,
                Index, or a dictionary of Series/DataFrames.
            as_pandas: If True, the folds are returned as a DataFrame. This is useful
                to visualize the folds in a more interpretable way. Defaults to False.

        Returns:
            A list of lists containing the indices (position) for each fold, or a
            DataFrame if `as_pandas=True`. Each list contains 4 lists and a boolean
            with the following information:

            - **fold**: fold number.
            - **[train_start, train_end]**: list with the start and end positions of
                    the training set.
            - **[last_window_start, last_window_end]**: list with the start and end
                    positions of the last window seen by the forecaster during training.
                    The last window is used to generate the lags use as predictors. If
                    `differentiation` is included, the interval is extended as many
                    observations as the differentiation order. If the argument `window_size`
                    is `None`, this list is empty.
            - **[test_start, test_end]**: list with the start and end positions of
                    the test set. These are the observations used to evaluate the forecaster.
            - **[test_start_with_gap, test_end_with_gap]**: list with the start and
                    end positions of the test set including the gap. The gap is the number
                    of observations between the end of the training set and the start of
                    the test set.
            - **fit_forecaster**: boolean indicating whether the forecaster should be
                    fitted in this fold.

        Note:
            The returned values are the positions of the observations and not the
            actual values of the index, so they can be used to slice the data directly
            using iloc.

            If `as_pandas` is `True`, the folds are returned as a DataFrame with the
            following columns: 'fold', 'train_start', 'train_end', 'last_window_start',
            'last_window_end', 'test_start', 'test_end', 'test_start_with_gap',
            'test_end_with_gap', 'fit_forecaster'.

            Following the python convention, the start index is inclusive and the end
            index is exclusive. This means that the last index is not included in the
            slice.
        """

        if not isinstance(X, (pd.Series, pd.DataFrame, pd.Index, dict)):
            raise TypeError(
                f"X must be a pandas Series, DataFrame, Index or a dictionary. "
                f"Got {type(X)}."
            )

        window_size_as_date_offset = isinstance(
            self.window_size, pd.tseries.offsets.DateOffset
        )
        if window_size_as_date_offset:
            # Calculate the window_size in steps. This is not a exact calculation
            # because the offset follows the calendar rules and the distance between
            # two dates may not be constant.
            first_valid_index = X.index[-1] - self.window_size
            try:
                window_size_idx_start = X.index.get_loc(first_valid_index)
                window_size_idx_end = X.index.get_loc(X.index[-1])
                self.window_size = window_size_idx_end - window_size_idx_start
            except KeyError:
                raise ValueError(
                    f"The length of `y` ({len(X)}), must be greater than or equal "
                    f"to the window size ({self.window_size}). This is because  "
                    f"the offset (forecaster.offset) is larger than the available "
                    f"data. Try to decrease the size of the offset (forecaster.offset), "
                    f"the number of `n_offsets` (forecaster.n_offsets) or increase the "
                    f"size of `y`."
                )

        if self.initial_train_size is None:
            if self.window_size is None:
                raise ValueError(
                    "To use split method when `initial_train_size` is None, "
                    "`window_size` must be an integer greater than 0. "
                    "Although no initial training is done and all data is used to "
                    "evaluate the model, the first `window_size` observations are "
                    "needed to create the initial predictors. Got `window_size` = None."
                )
            if self.refit:
                raise ValueError(
                    "`refit` is only allowed when `initial_train_size` is not `None`. "
                    "Set `refit` to `False` if you want to use `initial_train_size = None`."
                )
            externally_fitted = True
            self.initial_train_size = self.window_size  # Reset to None later
        else:
            if self.window_size is None:
                warnings.warn(
                    "Last window cannot be calculated because `window_size` is None.",
                    IgnoredArgumentWarning,
                )
            externally_fitted = False

        index = self._extract_index(X)
        idx = range(len(index))
        folds = []
        i = 0

        self.initial_train_size = date_to_index_position(
            index=index,
            date_input=self.initial_train_size,
            method="validation",
            date_literal="initial_train_size",
        )

        if window_size_as_date_offset:
            if self.initial_train_size is not None:
                if self.initial_train_size < self.window_size:
                    raise ValueError(
                        f"If `initial_train_size` is an integer, it must be greater than "
                        f"the `window_size` of the forecaster ({self.window_size}) "
                        f"and smaller than the length of the series ({len(X)}). If "
                        f"it is a date, it must be within this range of the index."
                    )

        if self.allow_incomplete_fold:
            # At least one observation after the gap to allow incomplete fold
            if len(index) <= self.initial_train_size + self.gap:
                raise ValueError(
                    f"The time series must have more than `initial_train_size + gap` "
                    f"observations to create at least one fold.\n"
                    f"    Time series length: {len(index)}\n"
                    f"    Required > {self.initial_train_size + self.gap}\n"
                    f"    initial_train_size: {self.initial_train_size}\n"
                    f"    gap: {self.gap}\n"
                )
        else:
            # At least one complete fold
            if len(index) < self.initial_train_size + self.gap + self.steps:
                raise ValueError(
                    f"The time series must have at least `initial_train_size + gap + steps` "
                    f"observations to create a minimum of one complete fold "
                    f"(allow_incomplete_fold=False).\n"
                    f"    Time series length: {len(index)}\n"
                    f"    Required >= {self.initial_train_size + self.gap + self.steps}\n"
                    f"    initial_train_size: {self.initial_train_size}\n"
                    f"    gap: {self.gap}\n"
                    f"    steps: {self.steps}\n"
                )

        while self.initial_train_size + (i * self.fold_stride) + self.gap < len(index):

            if self.refit:
                # NOTE: If `fixed_train_size` the train size doesn't increase but
                # moves by `fold_stride` positions in each iteration. If `False`,
                # the train size increases by `fold_stride` in each iteration.
                train_iloc_start = (
                    i * (self.fold_stride) if self.fixed_train_size else 0
                )
                train_iloc_end = self.initial_train_size + i * (self.fold_stride)
                test_iloc_start = train_iloc_end
            else:
                # NOTE: The train size doesn't increase and doesn't move.
                train_iloc_start = 0
                train_iloc_end = self.initial_train_size
                test_iloc_start = self.initial_train_size + i * (self.fold_stride)

            if self.window_size is not None:
                last_window_iloc_start = test_iloc_start - self.window_size

            test_iloc_end = test_iloc_start + self.gap + self.steps

            partitions = [
                idx[train_iloc_start:train_iloc_end],
                (
                    idx[last_window_iloc_start:test_iloc_start]
                    if self.window_size is not None
                    else []
                ),
                idx[test_iloc_start:test_iloc_end],
                idx[test_iloc_start + self.gap : test_iloc_end],
            ]
            folds.append(partitions)
            i += 1

        # NOTE: Delete all incomplete folds at the end if not allowed
        n_removed_folds = 0
        if not self.allow_incomplete_fold:
            # NOTE: While folds and the last "test_index_with_gap" is incomplete,
            # calculating len of range objects
            while folds and len(folds[-1][3]) < self.steps:
                folds.pop()
                n_removed_folds += 1

        # Replace partitions inside folds with length 0 with `None`
        folds = [
            [partition if len(partition) > 0 else None for partition in fold]
            for fold in folds
        ]

        # Create a flag to know whether to train the forecaster
        if self.refit == 0:
            self.refit = False

        if isinstance(self.refit, bool):
            fit_forecaster = [self.refit] * len(folds)
            fit_forecaster[0] = True
        else:
            fit_forecaster = [False] * len(folds)
            for i in range(0, len(fit_forecaster), self.refit):
                fit_forecaster[i] = True

        for i in range(len(folds)):
            folds[i].insert(0, i)
            folds[i].append(fit_forecaster[i])
            if fit_forecaster[i] is False:
                folds[i][1] = folds[i - 1][1]

        index_to_skip = []
        if self.skip_folds is not None:
            if isinstance(self.skip_folds, (int, np.integer)) and self.skip_folds > 0:
                index_to_keep = np.arange(0, len(folds), self.skip_folds)
                index_to_skip = np.setdiff1d(
                    np.arange(0, len(folds)), index_to_keep, assume_unique=True
                )
                index_to_skip = [
                    int(x) for x in index_to_skip
                ]  # Required since numpy 2.0
            if isinstance(self.skip_folds, list):
                index_to_skip = [i for i in self.skip_folds if i < len(folds)]

        if self.verbose:
            self._print_info(
                index=index,
                folds=folds,
                externally_fitted=externally_fitted,
                n_removed_folds=n_removed_folds,
                index_to_skip=index_to_skip,
            )

        folds = [fold for i, fold in enumerate(folds) if i not in index_to_skip]
        if not self.return_all_indexes:
            # NOTE: +1 to prevent iloc pandas from deleting the last observation
            folds = [
                [
                    fold[0],
                    [fold[1][0], fold[1][-1] + 1],
                    (
                        [fold[2][0], fold[2][-1] + 1]
                        if self.window_size is not None
                        else []
                    ),
                    [fold[3][0], fold[3][-1] + 1],
                    [fold[4][0], fold[4][-1] + 1],
                    fold[5],
                ]
                for fold in folds
            ]

        if externally_fitted:
            self.initial_train_size = None
            folds[0][5] = False

        if as_pandas:
            if self.window_size is None:
                for fold in folds:
                    fold[2] = [None, None]

            if not self.return_all_indexes:
                folds = pd.DataFrame(
                    data=[
                        [fold[0]] + list(itertools.chain(*fold[1:-1])) + [fold[-1]]
                        for fold in folds
                    ],
                    columns=[
                        "fold",
                        "train_start",
                        "train_end",
                        "last_window_start",
                        "last_window_end",
                        "test_start",
                        "test_end",
                        "test_start_with_gap",
                        "test_end_with_gap",
                        "fit_forecaster",
                    ],
                )
            else:
                folds = pd.DataFrame(
                    data=folds,
                    columns=[
                        "fold",
                        "train_index",
                        "last_window_index",
                        "test_index",
                        "test_index_with_gap",
                        "fit_forecaster",
                    ],
                )

        return folds

    def _print_info(
        self,
        index: pd.Index,
        folds: list[list[int]],
        externally_fitted: bool,
        n_removed_folds: int,
        index_to_skip: list[int],
    ) -> None:
        """Print information about folds.

        Args:
            index: Index of the time series data.
            folds: A list of lists containing the indices (position) for each fold.
            externally_fitted: Whether an already trained forecaster is to be used.
            n_removed_folds: Number of folds removed.
            index_to_skip: Number of folds skipped.
        """

        print("Information of folds")
        print("--------------------")
        if externally_fitted:
            print(
                f"An already trained forecaster is to be used. Window size: "
                f"{self.window_size}"
            )
        else:
            if self.differentiation is None:
                print(
                    f"Number of observations used for initial training: "
                    f"{self.initial_train_size}"
                )
            else:
                print(
                    f"Number of observations used for initial training: "
                    f"{self.initial_train_size - self.differentiation}"
                )
                print(
                    f"    First {self.differentiation} observation/s in training sets "
                    f"are used for differentiation"
                )
        print(
            f"Number of observations used for backtesting: "
            f"{len(index) - self.initial_train_size}"
        )
        print(f"    Number of folds: {len(folds)}")
        print(
            f"    Number skipped folds: "
            f"{len(index_to_skip)} {index_to_skip if index_to_skip else ''}"
        )
        print(f"    Number of steps per fold: {self.steps}")
        if self.steps != self.fold_stride:
            print(
                f"    Number of steps to the next fold (fold stride): {self.fold_stride}"
            )
        print(
            f"    Number of steps to exclude between last observed data "
            f"(last window) and predictions (gap): {self.gap}"
        )
        if n_removed_folds > 0:
            print(
                f"    The last {n_removed_folds} fold(s) have been excluded "
                f"because they were incomplete."
            )

        if len(folds[-1][4]) < self.steps:
            print(f"    Last fold only includes {len(folds[-1][4])} observations.")

        print("")

        if self.differentiation is None:
            differentiation = 0
        else:
            differentiation = self.differentiation

        for i, fold in enumerate(folds):
            is_fold_skipped = i in index_to_skip
            has_training = fold[-1] if i != 0 else True
            training_start = (
                index[fold[1][0] + differentiation] if fold[1] is not None else None
            )
            training_end = index[fold[1][-1]] if fold[1] is not None else None
            training_length = (
                len(fold[1]) - differentiation if fold[1] is not None else 0
            )
            validation_start = index[fold[4][0]]
            validation_end = index[fold[4][-1]]
            validation_length = len(fold[4])

            print(f"Fold: {i}")
            if is_fold_skipped:
                print("    Fold skipped")
            elif not externally_fitted and has_training:
                print(
                    f"    Training:   {training_start} -- {training_end}  "
                    f"(n={training_length})"
                )
                print(
                    f"    Validation: {validation_start} -- {validation_end}  "
                    f"(n={validation_length})"
                )
            else:
                print("    Training:   No training in this fold")
                print(
                    f"    Validation: {validation_start} -- {validation_end}  "
                    f"(n={validation_length})"
                )

        print("")



# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Manager module for model persistence, prediction, and training.

This module provides utilities for:
- Logging configuration for safety-critical systems
- Model persistence (save/load)
- Model prediction management
- Model training and retraining workflows
- Model evaluation metrics
- Dataset configurations (see manager.datasets submodule)
- CLI argument parsing utilities
"""

from spotforecast2_safe.manager.logger import setup_logging
from spotforecast2_safe.manager.metrics import calculate_metrics
from spotforecast2_safe.manager.persistence import (
    _ensure_model_dir,
    _get_model_filepath,
    _load_forecasters,
    _model_directory_exists,
    _save_forecasters,
)
from spotforecast2_safe.manager.predictor import get_model_prediction
from spotforecast2_safe.manager.tools import _parse_bool
from spotforecast2_safe.manager.trainer import get_last_model

__all__ = [
    # Logger
    "setup_logging",
    # Metrics
    "calculate_metrics",
    # Persistence
    "_ensure_model_dir",
    "_get_model_filepath",
    "_save_forecasters",
    "_load_forecasters",
    "_model_directory_exists",
    # Predictor
    "get_model_prediction",
    # Trainer
    "get_last_model",
    # Tools
    "_parse_bool",
]



# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later

from pathlib import Path
from typing import Dict, List, Tuple, Union

from joblib import dump, load

# ============================================================================
# Model Persistence Functions
# ============================================================================


def _ensure_model_dir(model_dir: Union[str, Path]) -> Path:
    """Ensure model directory exists.

    Args:
        model_dir: Directory path for model storage.

    Returns:
        Path: Validated Path object.

    Raises:
        OSError: If directory cannot be created.

    Examples:
        >>> import tempfile
        >>> from pathlib import Path
        >>> from spotforecast2_safe.manager.persistence import _ensure_model_dir
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     model_path = _ensure_model_dir(Path(tmpdir) / "models")
        ...     print(model_path.exists())
        True
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     nested = _ensure_model_dir(Path(tmpdir) / "a" / "b" / "c")
        ...     print(nested.is_dir())
        True
    """
    model_path = Path(model_dir)
    model_path.mkdir(parents=True, exist_ok=True)
    return model_path


def _get_model_filepath(model_dir: Path, target: str) -> Path:
    """Get filepath for a single model.

    Args:
        model_dir: Directory containing models.
        target: Target variable name.

    Returns:
        Path: Full filepath for the model.

    Examples:
        >>> from pathlib import Path
        >>> from spotforecast2_safe.manager.persistence import _get_model_filepath
        >>> path = _get_model_filepath(Path("./models"), "power")
        >>> str(path)
        'models/forecaster_power.joblib'
        >>> path = _get_model_filepath(Path("/tmp/models"), "energy")
        >>> path.name
        'forecaster_energy.joblib'
        >>> path.suffix
        '.joblib'
    """
    return model_dir / f"forecaster_{target}.joblib"


def _save_forecasters(
    forecasters: Dict[str, object],
    model_dir: Union[str, Path],
    verbose: bool = False,
) -> Dict[str, Path]:
    """Save trained forecasters to disk using joblib.

    Follows scikit-learn persistence conventions using joblib for efficient
    serialization of sklearn-compatible estimators.

    Args:
        forecasters: Dictionary mapping target names to trained ForecasterEquivalentDate objects.
        model_dir: Directory to save models. Created if it doesn't exist.
        verbose: Print progress messages. Default: False.

    Returns:
        Dict[str, Path]: Dictionary mapping target names to saved model filepaths.

    Raises:
        OSError: If models cannot be written to disk.
        TypeError: If forecasters contain non-serializable objects.

    Examples:
        >>> import tempfile
        >>> from pathlib import Path
        >>> from spotforecast2_safe.manager.persistence import _save_forecasters
        >>> from sklearn.linear_model import LinearRegression
        >>> mock_model = LinearRegression()
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     forecasters = {"power": mock_model, "energy": mock_model}
        ...     paths = _save_forecasters(forecasters, tmpdir, verbose=False)
        ...     print("power" in paths)
        ...     print(paths["power"].exists())
        True
        True
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     paths = _save_forecasters({"demand": mock_model}, tmpdir)
        ...     print(paths["demand"].suffix)
        .joblib
    """
    model_path = _ensure_model_dir(model_dir)
    saved_paths = {}

    for target, forecaster in forecasters.items():
        filepath = _get_model_filepath(model_path, target)
        try:
            dump(forecaster, filepath, compress=3)
            saved_paths[target] = filepath
            if verbose:
                print(f"  ✓ Saved forecaster for {target} to {filepath}")
        except Exception as e:
            raise OSError(f"Failed to save model for {target}: {e}")

    return saved_paths


def _load_forecasters(
    target_columns: List[str],
    model_dir: Union[str, Path],
    verbose: bool = False,
) -> Tuple[Dict[str, object], List[str]]:
    """Load trained forecasters from disk using joblib.

    Attempts to load all forecasters for given targets. Missing models
    are indicated in the return value for selective retraining.

    Args:
        target_columns: List of target variable names to load.
        model_dir: Directory containing saved models.
        verbose: Print progress messages. Default: False.

    Returns:
        Tuple[Dict[str, object], List[str]]:
        - forecasters: Dictionary of successfully loaded ForecasterEquivalentDate objects.
        - missing_targets: List of target names without saved models.

    Examples:
        >>> import tempfile
        >>> from pathlib import Path
        >>> from spotforecast2_safe.manager.persistence import (
        ...     _save_forecasters,
        ...     _load_forecasters,
        ... )
        >>> from sklearn.linear_model import LinearRegression
        >>> mock_model = LinearRegression()
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     _ = _save_forecasters({"power": mock_model}, tmpdir)
        ...     forecasters, missing = _load_forecasters(
        ...         ["power", "energy"],
        ...         tmpdir,
        ...         verbose=False
        ...     )
        ...     print("power" in forecasters)
        ...     print("energy" in missing)
        True
        True
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     forecasters, missing = _load_forecasters(["nonexistent"], tmpdir)
        ...     print(len(forecasters), len(missing))
        0 1
    """
    model_path = Path(model_dir)
    forecasters = {}
    missing_targets = []

    for target in target_columns:
        filepath = _get_model_filepath(model_path, target)
        if filepath.exists():
            try:
                forecasters[target] = load(filepath)
                if verbose:
                    print(f"  ✓ Loaded forecaster for {target} from {filepath}")
            except Exception as e:
                if verbose:
                    print(f"  ✗ Failed to load {target}: {e}")
                missing_targets.append(target)
        else:
            missing_targets.append(target)

    return forecasters, missing_targets


def _model_directory_exists(model_dir: Union[str, Path]) -> bool:
    """Check if model directory exists.

    Args:
        model_dir: Directory path to check.

    Returns:
        bool: True if directory exists, False otherwise.

    Examples:
        >>> import tempfile
        >>> from pathlib import Path
        >>> from spotforecast2_safe.manager.persistence import _model_directory_exists
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     print(_model_directory_exists(tmpdir))
        True
        >>> print(_model_directory_exists("/nonexistent/path/to/models"))
        False
    """
    return Path(model_dir).exists()
